% File 'cal.m'
% 
% For GNU Octave
% --------------------------------------------------------------------------------------------------
% --------------------------------------------------------------------------------------------------
% 
% Octave script to perform a calibration of the Mk III Signal Generator.
%
% This script is written to work with a Rohde & Schwarz NRVD power meter
% with power sensor connected to channel A. The power meter is assumed to 
% be controlled by a Prologix GPIB/USB controller.
% 
% Author: Dr. Mario Hellmich <mario.hellmich@web.de>
% Date:   10.07.2023
% 
% Uses the Octave Forge 'instrument-control' package 
%
% --------------------------------------------------------------------------------------------------
% --------------------------------------------------------------------------------------------------



1;                                                                           				   		% Prevent Octave from thinking this is a function file
clear -all;                                                                        					% Delete all
clc;																								% Clear screen
pkg load "instrument-control"																		% Load instrument control package
pause(0.2);


% --------------------------------------------------------------------------------------------------
% ----- Com Setup ----------------------------------------------------------------------------------

global instr = serialport("COM9", 9600);															% Open virtual COM port for the signal generator (must be the port of the signal generator)
global meter = serialport("COM6", 9600);															% Open virtual COM port for the power meter (must be the port of the GPIB controller connected to the R&S NRVD)

set(instr, 'DataBits', 8);
set(instr, 'parity', 'n');
set(instr, 'stopbits', 1);
set(instr, 'timeout', 10);
configureTerminator(instr, 'lf', 'lf');

set(meter, 'DataBits', 8);
set(meter, 'parity', 'n');
set(meter, 'stopbits', 1);
set(meter, 'timeout', 10);
configureTerminator(meter, 'lf', 'lf');

flush(instr);																						% Flush port
flush(meter);																						% Flush port


% --------------------------------------------------------------------------------------------------
% ----- Constants ----------------------------------------------------------------------------------

global address = 13;																				% GPIB address of my R&S NRVD
global password = "123456"																			% Signal generator calibration unlock password

global fmin = 1e6;																					% Minimum frequency (Hz)
global fmax = 4.2e9;																				% Maximum frequency (Hz)
global ftrans = 100e6;																				% Transition between coarse and fine calibration value spacing (Hz)
global spacing_lo = 5e6;																			% Frequency spacing of level correction values (Hz) below ftrans
global spacing_hi = 100e6;																			% Frequency spacing of level correction values (Hz) above ftrans
global natt_stages = 4;																				% Number of attenuator stages
global db_step = 3;																					% Minimal attenuator step (dB)
global nominal_lvl_pre_att = 0;																		% Nominal power level before the attenuator (dBm)

global ref_freq = 10e6;																				% Signal generator reference frequency (Hz)
global ref_corr_max = 1000;																			% Maximum frequency correction value
global ref_corr_min = -1000;																		% Minimum frequency correction value

global lvl_corr_max = 50;																			% Maximum level correction value
global lvl_corr_min = -50;																			% Minimum level correction value
global lvl_precision = 0.025;																		% Exit condition of cal algorithm for level deviation (dB)
global lvl_maxiter = 25;																			% Exit condition of cal algorithm for number of iterations

global slp_corrfrq = 1e9;																			% Frequency at which slope correction is performed
global slp_corrlvl_diff = 5;																		% Level difference between slope correction high and low level setting (dB)
global slp_pwr_high = nominal_lvl_pre_att + slp_corrlvl_diff;										% Power at slope correction high level setting (dBm)
global slp_pwr_low = nominal_lvl_pre_att;															% Power at slope correction low level setting (dBm)
global slp_corr_max = 300;																			% Maximum slope correction value
global slp_corr_min = -300;																			% Minimum slope correction value
global slp_corr_hi_init = 3;																		% Slope correction start value
global slp_corr_lo_init = -3;																		% Slope correction start value
global slp_precision = 0.01;																		% Exit condition of cal algorithm for detector slope deviation (dB)
global slp_maxiter = 8;																				% Exit condition of cal algorithm for number of iterations


% --------------------------------------------------------------------------------------------------
% ----- Utility functions --------------------------------------------------------------------------

function int_array = readtolf(ser_handle)															% Read characters from a serial port with handle ser_handle until a line-feed character occurs
	int_array = [];
	do
		nextchar = fread(ser_handle, 1);
		int_array(end + 1) = nextchar;
	until(char(nextchar) == "\n")
endfunction


function retval = stage_to_dbm(att_stage)															% Returns the nominal output power in dBm in calibration mode for selected attenuator stage att_stage
	global nominal_lvl_pre_att;
	global db_step;
	if(att_stage == 0)
		retval = nominal_lvl_pre_att;
	else
		retval = nominal_lvl_pre_att - (db_step * 2^(att_stage - 1));
	endif
endfunction


function retval = lvl_error(att_stage, corr) 														% Returns the error in dB between measured and nominal output level when att_stage is engaged and correction value corr is set
	global instr;
	global meter;
	fwrite(meter, "*CLS\n");																		% Clear output buffer of meter
	flush(meter);
	fwrite(instr, ["cal:att ", num2str(att_stage), "\r\n"]);										% Set the attenuator stage to calibrate
	fwrite(instr, ["cal:lvl ", num2str(corr, 6), "\r\n"]);											% Set signal generator to corr value
	pause(0.5);																						% Wait for output to settle
	fwrite(meter, "*TRG;*WAI\n");																	% Trigger NRVD
	meas = str2num(char(readtolf(meter)));															% Measure and convert level
	retval = meas - stage_to_dbm(att_stage);														% Calculate error and return
endfunction


function retval = slp_error(corr)																	% Returns the error in dB between measured and nominal level differences (high and low power level) for detector slope correction
	global instr;
	global meter;
	global slp_pwr_high;
	global slp_pwr_low;
	fwrite(meter, "*CLS\n");																		% Clear output buffer of meter
	flush(meter);
	fwrite(instr, ["cal:scr ", num2str(corr, 6), "\r\n"]);											% Set signal generator to corr value
	fwrite(instr, "cal:mde 1\r\n");																	% Set high level mode
	pause(0.5);
	fwrite(meter, "*TRG;*WAI\n");
	meas_high = str2num(char(readtolf(meter)));
	fwrite(instr, "cal:mde 2\r\n");																	% Set low level mode
	pause(0.5);
	fwrite(meter, "*TRG;*WAI\n");
	meas_low = str2num(char(readtolf(meter)));
	retval = (meas_high - meas_low) - (slp_pwr_high - slp_pwr_low);
endfunction


% --------------------------------------------------------------------------------------------------
% ----- Setup --------------------------------------------------------------------------------------

printf("-------------------------------------------------------------\n");
printf("Automatic Calibration of the MK-III Signal Generator\n");
printf("-------------------------------------------------------------\n");
printf("Resetting equipment, please wait...\n\n");

errorflag = false;																					% Set to true when an error has occurred

fwrite(meter, "++mode 1\n");																		% Set the GPIB controller to controller mode
%fwrite(meter, ["++addr ", num2str(address), " \r\n"]);												% Set NRVD GPIB address (stored by the Prologix adapter)
%fwrite(meter, "++auto 1\n");																		% Set GPIB controller to auto listen mode

fwrite(meter, "*RST;*CLS;\n");																		% NRVD reset and clear status
fwrite(meter, "INP:SEL \"B\"\n");																	% Select channel B
fwrite(meter, "SENS:POW:UNIT DBM\n");																% Select power measurement and unit dBm
fwrite(meter, "DISP:ANN:POW:NRES 5\n");																% Select 5 digits resolution
fwrite(meter, "DISP:ANN:BARG ON\n");																% Switch bargraph on
fwrite(meter, "DISP:ANN:BARG:AUTO\n");																% Bargraph autoranging
fwrite(meter, "SENS:CORR:FREF:STAT ON\n");															% Frequency response correction on
fwrite(meter, "DIAG:TEST2 OFF\n");																	% No default string when output buffer is empty

fwrite(instr, "rst\r\n");																			% Reset signal generator
pause(4);																							% Wait for the signal generator to reset
fwrite(instr, ["cal:on ", password, " \r\n"]);														% Select calibration mode of signal generator

sel = yes_or_no("Zeroing the power sensor is recommended. Do you want to zero? ");
if(sel == true)
	disp("Zeroing power sensor. Disconnect sensor from signal generator and press any key when ready.")
	kbhit();
	fwrite(meter, "SENS:CORR:ZERO:INIT\n");															% Zero NRVD
	pause(3);
	disp("When zeroing is finished, connect power sensor to signal generator and press any key when ready to calibrate.");
	kbhit();
else
	disp("Connect power sensor to signal generator and press any key when ready to calibrate.");
	kbhit();
endif


% --------------------------------------------------------------------------------------------------
% ----- Detector slope correction ------------------------------------------------------------------

fwrite(instr, ["cw:frq ", num2str(slp_corrfrq), " Hz\r\n"]);										% Set signal generator to required frequency
fwrite(instr, "cal:scr 0.0\r\n");																	% Initially, set slope correction value to 0.
fwrite(meter, ["SENS:CORR:FREF ", num2str(slp_corrfrq), " Hz\n"]);									% Set power meter level correction to required frequency

disp("\nCalibrating detector slope correction...");

len = slp_corr_hi_init - slp_corr_lo_init;
corr_v = linspace(slp_corr_lo_init, slp_corr_hi_init, 11);
err_v  = arrayfun(@slp_error, corr_v);
[curr_err, idx] = min(abs(err_v));																	% Error of current correction value
curr_corr = corr_v(idx);																			% Current correction value
iter = 0;																							% Iteration count

while(true)
	iter++;
	printf("Iteration %d, current error %e.\n", iter, curr_err);
	if( abs(curr_err) <= slp_precision )
		printf("Success: Slope error %e dB over a %d dB level step after %d iteration(s), correction value %e.\n", curr_err, slp_pwr_high - slp_pwr_low, iter, curr_corr);
		fwrite(instr, ["cal:scr ", num2str(curr_corr, 6), "\r\n"]);
		break;
	elseif(iter >= slp_maxiter)
		printf("Failure: Max. number of iterations reached, slope error %e dB over a %d dB level step.\n", curr_err, slp_pwr_high - slp_pwr_low);
		fwrite(instr, ["cal:scr ", num2str(curr_corr, 6), "\r\n"]);
		errorflag = true;
		break;
	else
		low  = max(slp_corr_lo_init, curr_corr - len/6);
		high = min(slp_corr_hi_init, curr_corr + len/6);
		len = high - low;
		corr_v = linspace(low, high, 11);
		err_v  = arrayfun(@slp_error, corr_v);
		[curr_err, idx] = min(abs(err_v));
		curr_corr = corr_v(idx);
	endif
endwhile


% --------------------------------------------------------------------------------------------------
% ----- Level correction ---------------------------------------------------------------------------

freq = [[fmin:spacing_lo:ftrans], [ftrans:spacing_hi:fmax]];										% Create array with calibration frequencies

fwrite(instr, "cal:mde 0\r\n");																		% Select calibration mode 0

disp("\nCalibrating output level and attenuator...");

for att = (0:natt_stages)
	for f = freq
		fwrite(instr, ["cw:frq ", num2str(f), " Hz\r\n"]);											% Set signal generator to new frequency
		fwrite(meter, ["SENS:CORR:FREF ", num2str(f), " Hz\n"]);									% Set NRVD power sensor correction frequency
		
		printf("\r\nCalibrating attenuator stage %d at %e Hz:\n", att, f);

		curr_corr = -lvl_error(att, 0);																% Current correction value
		curr_err = lvl_error(att, curr_corr);														% Error of current correction value
		iter = 0;																					% Iteration count
		
		while(true)
			iter++;
			if( abs(curr_err) <= lvl_precision )
				printf("Success: Stage %d at %e Hz with %e dB error after %d iteration(s), correction value %e dB.\n", att, f, curr_err, iter, curr_corr);
				fwrite(instr, ["cal:lvl ", num2str(curr_corr, 6), "\r\n"]);
				break;
			elseif(iter >= lvl_maxiter)
				printf("Failure: Stage %d at %e Hz after max. number of iterations reached, error %e dB, correction value %e dB.\n", att, f, curr_err, curr_corr);
				fwrite(instr, ["cal:lvl ", num2str(curr_corr, 6), "\r\n"]);
				errorflag = true;
				break;
			elseif( (curr_corr > lvl_corr_max) || (curr_corr < lvl_corr_min) )
				printf("Failure: Stage %d at %e Hz after exceeding maximum or minimum correction value, error %e dB.\n", att, f, curr_err);
				if(curr_corr > lvl_corr_max)
					fwrite(instr, ["cal:lvl ", num2str(lvl_corr_max, 6), "\r\n"]);
				else
					fwrite(instr, ["cal:lvl ", num2str(lvl_corr_min, 6), "\r\n"]);
				endif
				errorflag = true;
				break;
			else
				curr_corr = curr_corr - (0.9 * curr_err);
				curr_err = lvl_error(att, curr_corr);
			endif
		endwhile;
	endfor
endfor


% --------------------------------------------------------------------------------------------------
% ----- Time base correction -----------------------------------------------------------------------

disp("\nTimebase calibration...");
disp("Connect signal generator to a frequency counter and press any key when ready.");
kbhit();
fwrite(instr, "cal:mde 0\r\n");
fwrite(instr, "cal:att 0\r\n");
fwrite(instr, ["cw:frq ", num2str(ref_freq), " Hz\r\n"]);
do
	fwrite(instr, "cal:frq 0\r\n");
	meas_freq = input("Enter measured frequency in Hz: ");
	freq_corr = meas_freq - ref_freq;
	if( (freq_corr <= ref_corr_max) || (freq_corr >= ref_corr_min) )
		fwrite(instr, ["cal:frq ", num2str(freq_corr, 16), "\r\n"]);
	else
		disp("Correction out of bounds.");
	endif
until( yes_or_no("Repeat? ") == false );	


% --------------------------------------------------------------------------------------------------
% ----- Finishing: save and reset signal generator -------------------------------------------------

if(errorflag == false)
	disp("\nCalibration finished without errors.\n");
else
	disp("\nCalibration finished with errors.\n");
endif

fwrite(instr, ["cal:cst Cal-Date ", date(), "\r\n"]);												% Write calibration string (current date)

sel = yes_or_no("Save calibration data to EEPROM? ");
if(sel == true)
	fwrite(instr, "cal:sve\r\n");																	% Save calibration data in EEPROM
endif
fwrite(instr, "rst\r\n");																			% Reset signal generator to leave calibration mode
fwrite(meter, "*GTL\n");																			% NRVD go to local
fclose("all");																						% Close all ports

% --------------------------------------------------------------------------------------------------
% End of File
% 