/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file implements the logic of the console application's 
 * state machine.
 * 
 * *****************************************************************************/

/*Section: Includes ************************************************************/

#include "app.h"


/*Section: Definitions *********************************************************/

static APP_STATES appState;

#define APP_CDC_BUF_SZZE        64

static uint8_t CACHE_ALIGN cdc_rx_buffer[APP_CDC_BUF_SZZE];
static uint8_t CACHE_ALIGN cdc_tx_buffer[APP_CDC_BUF_SZZE];


/*Section: Function implementations ********************************************/

void APP_Initialize(void)
{
    appState = APP_STATE_INIT;
}


void APP_Tasks(void)
{
    switch(appState)
    {
        case APP_STATE_INIT:
            /* Initialize sweep timer and load default state */
            Sweeptimer_Init();
            set_state_to_default(&state, &error);
            
            /* Reset RTOS queues */
            xQueueReset(cdc_rx_queue);
            xQueueReset(cdc_tx_queue);
            xQueueReset(spi_queue);
            
            /* Flush console buffer */
            SYS_CONSOLE_Flush(SYS_CONSOLE_DEFAULT_INSTANCE);
            
            /* Initialize synthesizer hardware */
            HW_Initialize();
                        
            /* Load calibration data into RAM, recall memory 0 and set up 
             * instrument to recalled state */
            load_cal_data();                                                    // Load calibration data; should be called after set_state_to_default() so that a checksum error is not immediately cleared
            mem_recall(0);                                                      // Recall state from memory 0
            ref_update();                                                       // Update reference
            /* synthesizer_update() is already called in ref_update() */
            
            appState = APP_STATE_READ_DATA;
            break;
            
            
        case APP_STATE_READ_DATA:   ;
            /* Check if new data is in USB CDC receive buffer, and put it into the FIFO queue */
            unsigned int bytes_in_buffer = SYS_CONSOLE_ReadCountGet(SYS_CONSOLE_DEFAULT_INSTANCE);
            if(bytes_in_buffer > 0)
            {
                unsigned int nbytestoread = (bytes_in_buffer < APP_CDC_BUF_SZZE) ? bytes_in_buffer : APP_CDC_BUF_SZZE;
                unsigned int nbytesread = SYS_CONSOLE_Read(SYS_CONSOLE_DEFAULT_INSTANCE, cdc_rx_buffer, nbytestoread);
                for(int i = 0; i < nbytesread; i++)
                    xQueueSendToBack(cdc_rx_queue, cdc_rx_buffer + i, portMAX_DELAY);
            }
            appState = APP_STATE_WRITE_DATA;
            break;
            
            
        case APP_STATE_WRITE_DATA:  ;
            /* Check if there is any data to send, and transmit it over the USB CDC data link */
            unsigned int nbytescopied = 0;
            unsigned int nbytessent = 0;
            char next_char;
            while( (nbytescopied < APP_CDC_BUF_SZZE) && (xQueueReceive(cdc_tx_queue, &next_char, (TickType_t)1) == pdTRUE) )
                cdc_tx_buffer[nbytescopied++] = next_char;
            if(nbytescopied > 0)
            {
                do
                    nbytessent += SYS_CONSOLE_Write(SYS_CONSOLE_DEFAULT_INSTANCE, cdc_tx_buffer + nbytessent, nbytescopied - nbytessent);
                while(nbytessent < nbytescopied);
            }
            appState = APP_STATE_READ_DATA;
            break;
    }
}


/*******************************************************************************
 End of File
 */