/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app_cmd_parser.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file implements part of the communication over the USB CDC 
 * data link as well as the command parser. It identifies commands in the 
 * USB CDC data stream and stores them in a struct. This struct is then read 
 * and the commands are executed. If necessary, data is sent in return over 
 * the USB CDC data link.
 * 
 * Command lines have the following structure (where <.> means that the part is 
 * optional):
 * 
 *      PREFIX <numerical argument> <unit> <?> CR/LF 
 *      PREFIX:SUFFIX <numerical argument> <unit> <?> CR/LF/;
 * 
 * Numerical arguments may contain a decimal point and have a sign. Command 
 * lines are terminated by \r or \n (or \r\n). Queries are terminated by a 
 * question mark. They are answered by sending a string in return.
 * 
 * *****************************************************************************/

/*Section: Includes ************************************************************/

#include "app_cmd_parser.h"


/*Section: Type definitions ****************************************************/

typedef enum {
    /* No prefix found */
    PRF_VOID = 0,
    
    /* Get identification string */
    PRF_IDENTIFY,
    
    /* Get firmware version */
    PRF_VERSION,
    
    /* Get current errors */
    PRF_ERROR,
    
    /* Get error status information */
    PRF_GET_STATUS,
            
    /* Reset instrument*/
    PRF_RESET,
            
    /* Clear status (clear all errors) */
    PRF_CLEAR_STATUS,
     
    /* Prefix for CW settings */
    PRF_CW,
    
    /* Prefix for frequency sweep settings */
    PRF_SWEEP,
    
    /* Set RF output on or off */
    PRF_RF_OUTPUT,
    
    /* Prefix for reference settings */
    PRF_REFERENCE,
            
    /* Save the current state in memory */
    PRF_SAVE,

    /* Recall a state from memory */
    PRF_RECALL,
            
    /* Prefix for calibration functions */
    PRF_CAL,
    
    /* Number of command literals */
    numPrefix
} Cmd_Prefix_t;


typedef enum {
    /* No suffix found */
    SUF_VOID = 0,
    
    /* Set or get current frequency */
    SUF_FREQUENCY,
    
    /* Set or get current output level */
    SUF_LEVEL,

    /* Set or get on or off status of a feature determined by prefix */
    SUF_SWITCH_ON,
    
    /* Set or get sweep start frequency */
    SUF_START_FREQUENCY,
    
    /* Set or get sweep stop frequency */
    SUF_STOP_FREQUENCY,
    
    /* Set or get sweep dwell time */
    SUF_DWELL_TIME,
    
    /* Set or get sweep number of points */
    SUF_NPOINTS,
    
    /* Set or get sweep single mode state */
    SUF_SINGLE,
    
    /* Set or get current reference status (0: internal, 1: external) */
    SUF_REF_EXTERNAL,
    
    /* Set or get calibration mode */
    SUF_CAL_MODE,
    
    /* Set or get attenuator stage for calibration */
    SUF_ATT_STAGE,
    
    /* Set or get ALC detector slope correction value */
    SUF_SLOPE_CORR,
    
    /* Set calibration string */
    SUF_CALIBRATION_STRING,
    
    /* Save calibration data in EEPROM */
    SUF_SAVE_CALDATA,
    
    /* Dump all calibration data from RAM over the USB CDC data link */
    SUF_DUMP_CALDATA,
    
    /* Erase EEPROM */
    SUF_ERASE_EEPROM,
    
    /* Number of command literals */
    numSuffix
} Cmd_Suffix_t;


typedef enum {
    UNIT_VOID = 0,
    UNIT_GHZ,
    UNIT_MHZ,
    UNIT_KHZ,
    UNIT_HZ,
    UNIT_DBM,
    UNIT_UV,
    UNIT_MV,
    UNIT_V,
    UNIT_US,
    UNIT_MS,
    UNIT_S,
    numUnit
} Cmd_Unit_t;


typedef struct {
    Cmd_Prefix_t prefix;
    Cmd_Suffix_t suffix;
    Cmd_Unit_t unit;
    bool query;
    uint64_t arg_integer;
    uint64_t arg_fractional;
    unsigned int lzeros;
    bool arg_sign;
    char string[STRARG_LEN + 1];
} Command_Parsed_t;


typedef enum {
    CMD_PARSER_READ_NEW_DATA = 0,
    CMD_PARSER_PARSE_COMMAND,
    CMD_PARSER_EXECUTE_COMMAND
} CMD_Parser_State_t;


/*Section: Local variables *****************************************************/

#define COMBUFFER_LEN      64

typedef struct {
    char CACHE_ALIGN str[COMBUFFER_LEN + 1];
    bool ready;
} Com_Buffer_t;

Com_Buffer_t local_read_buffer  = {.str = "", .ready = false};
Com_Buffer_t local_write_buffer = {.str = "", .ready = true};


/*Section: Function implementations ********************************************/

/*******************************************************
 * The following three functions return the corresponding 
 * command string literal (prefix, suffix and unit) for
 * an argument of type Cmd_Prefix_t, Cmd_Suffix_t and
 * Cmd_Unit_t.
 *******************************************************/
 
static const char *prefix_literal(Cmd_Prefix_t cmd)
{
    switch(cmd)
    {
        case PRF_IDENTIFY:              return "IDN";
        case PRF_VERSION:               return "VER";
        case PRF_ERROR:                 return "ERR";
        case PRF_GET_STATUS:            return "STS";
        case PRF_RESET:                 return "RST";
        case PRF_CLEAR_STATUS:          return "CLS";
        case PRF_CW:                    return "CW";
        case PRF_SWEEP:                 return "SWP";
        case PRF_RF_OUTPUT:             return "RFE";
        case PRF_REFERENCE:             return "REF";
        case PRF_SAVE:                  return "SVE";
        case PRF_RECALL:                return "RCL";
        case PRF_CAL:                   return "CAL";
        default:                        return "";
    }
}


static const char *suffix_literal(Cmd_Suffix_t cmd)
{
    switch(cmd)
    {
        case SUF_FREQUENCY:             return "FRQ";
        case SUF_LEVEL:                 return "LVL";
        case SUF_SWITCH_ON:             return "ON";
        case SUF_START_FREQUENCY:       return "STR";
        case SUF_STOP_FREQUENCY:        return "STP";
        case SUF_DWELL_TIME:            return "DWT";
        case SUF_NPOINTS:               return "NPT";
        case SUF_SINGLE:                return "SGL";
        case SUF_REF_EXTERNAL:          return "EXT";
        case SUF_ATT_STAGE:             return "ATT";
        case SUF_CAL_MODE:              return "MDE";
        case SUF_SLOPE_CORR:            return "SCR";
        case SUF_CALIBRATION_STRING:    return "CST";
        case SUF_SAVE_CALDATA:          return "SVE";
        case SUF_DUMP_CALDATA:          return "DMP";
        case SUF_ERASE_EEPROM:          return "ERS";
        default:                        return "";
    }
}


static const char *unit_literal(Cmd_Unit_t cmd)
{
    switch(cmd)
    {
        case UNIT_GHZ:                  return "GHZ";
        case UNIT_MHZ:                  return "MHZ";
        case UNIT_KHZ:                  return "KHZ";
        case UNIT_HZ:                   return "HZ";
        case UNIT_DBM:                  return "DBM";
        case UNIT_V:                    return "V";
        case UNIT_MV:                   return "MV";
        case UNIT_UV:                   return "UV";
        case UNIT_MS:                   return "MS";
        case UNIT_US:                   return "US";
        case UNIT_S:                    return "S";
        default:                        return "";
    }
}


/*******************************************************
 * The following functions update the local read and 
 * write buffers from the send and receive queues. 
 * Moreover, rx_buffer_update identifies command lines.
 *******************************************************/

static void rx_buffer_update(Com_Buffer_t *rx_buffer)
{
    static unsigned int nbytesread = 0;
    bool cmdline_complete = false;
    if(rx_buffer->ready == false)
    {
        char next_char = 0;
        while(xQueueReceive(cdc_rx_queue, &next_char, (TickType_t)0) == pdTRUE)
        {
            if((next_char != '\n') && (next_char != '\r') && (next_char != ';') && (nbytesread < COMBUFFER_LEN))            // Command line can be terminated by \n, \r or ';'
                rx_buffer->str[nbytesread++] = next_char;
            else
            {
                cmdline_complete = true;
                break;
            }
        }
        if(cmdline_complete == true)
        {
            rx_buffer->str[nbytesread] = '\0';
            nbytesread = 0;
            rx_buffer->ready = true;
        }
    }
}
 

static void tx_buffer_update(Com_Buffer_t *tx_buffer)
{
    if(tx_buffer->ready == true)
    {
        for(int i = 0; ( (tx_buffer->str[i] != '\0') && (i < COMBUFFER_LEN) ); i++)
            xQueueSendToBack(cdc_tx_queue, tx_buffer->str + i, (TickType_t)10);
        tx_buffer->ready = false;
    }
}


/*******************************************************
 * The following functions are used in command parsing. 
 * They serve to identify whether a command is a query,
 * to extract the argument and the unit, and to convert
 * ASCII to integer.
 *******************************************************/

static void strcpy_toupper(char *dest, const char *src)
{
    unsigned int i = 0;
    while(src[i] != '\0')
    {
        dest[i] = toupper(src[i]);
        i++;
    }
    dest[i] = '\0';
}


static Cmd_Prefix_t get_prefix(const char *str)
{
    if(str == NULL)
        return PRF_VOID;
    char buf[strlen(str) + 1];
    strcpy_toupper(buf, str);
    for(Cmd_Prefix_t itr = PRF_VOID + 1; itr < numPrefix; itr++)
    {
        const char *itr_str = prefix_literal(itr);
        if(itr_str[0] == '\0')
            return PRF_VOID;
        char *ptr = strstr(buf, itr_str);
        if(ptr != NULL)
            return itr;
    }
    return PRF_VOID;
}


static Cmd_Suffix_t get_suffix(const char *str)
{
	if(str == NULL)
        return SUF_VOID;
    char buf[strlen(str) + 1];
    strcpy_toupper(buf, str);
    for(Cmd_Suffix_t itr = SUF_VOID + 1; itr < numSuffix; itr++)
    {
        const char *itr_str = suffix_literal(itr);
        if(itr_str[0] == '\0')
            return SUF_VOID;
        char *ptr = strstr(buf, itr_str);
        if(ptr != NULL)
            return itr;
    }
    return SUF_VOID;
}


static Cmd_Unit_t get_unit(const char *str)
{
	if(str == NULL)
        return UNIT_VOID;
    char buf[strlen(str) + 1];
    strcpy_toupper(buf, str);
    for(Cmd_Unit_t itr = UNIT_VOID + 1; itr < numUnit; itr++)
    {
        const char *itr_str = unit_literal(itr);
        if(itr_str[0] == '\0')
            return UNIT_VOID;
        char *ptr = strstr(buf, itr_str);
        if(ptr != NULL)
            return itr;
    }
    return UNIT_VOID;
}


static void remove_space(char *str)
{
	if(str == NULL)
        return;
    int count = 0;
	for(int i = 0; str[i]; i++)
	{
		if( !isspace(str[i]) )
			str[count++] = str[i];
	}
	str[count] = '\0';
}


static void remove_ctrl(char *str)
{
    if(str == NULL)
        return;
    int count = 0;
    for(int i = 0; str[i]; i++)
    {
        if(isprint(str[i]) == true)
            str[count++] = str[i];
    }
    str[count] = '\0';
}


static bool is_query(const char *str)
{
	if(str == NULL)
        return false;
    char *ptr = strchr(str, '?');
    if(ptr == NULL)
        return false;
    else
        return true;
}


static void extract_digits(char *str)
{
	if(str == NULL)
        return;
    unsigned int count = 0;
	for(int i = 0; str[i]; i++)
	{
		if(isdigit(str[i]) || str[i] == '-' || str[i] == '.')
			str[count++] = str[i];
	}
	str[count] = '\0';
}


static unsigned int nleading_zeros(const char *str)
{
    unsigned int count = 0;
    for(int i = 0; (str[i] != '\0') && (str[i] == '0'); i++)
        count++;
    return count;
}


static void convert_num_arg(char *str, Command_Parsed_t *cmd)
{
    /* Set arguments to zero */
    cmd->arg_integer = 0;
    cmd->arg_fractional = 0;
    cmd->lzeros = 0;
    cmd->arg_sign = false;
    cmd->string[0] = '\0';
    
    if(str == NULL)
        return;
    
    /* Check if string starts with a decimal point because strtok does not handle this case */
    char *integer_tok, *fractional_tok;
    if(str[0] == '.')
    {
        integer_tok = "0";
        fractional_tok = str + 1;
    }
    else
    {
        integer_tok = strtok(str, ".");
        fractional_tok = strtok(NULL, ".");
    }

    /* Check for sign and store result */
    if(integer_tok[0] == '-')
    {
        cmd->arg_sign = true;
        integer_tok[0] = '0';
    }
    else
        cmd->arg_sign = false;

    /* Convert to numbers */
    cmd->arg_integer = strtoull(integer_tok, NULL, 10);
    cmd->lzeros = nleading_zeros(fractional_tok);
    cmd->arg_fractional = strtoull(fractional_tok, NULL, 10);
}


static void convert_str_arg(const char *str, Command_Parsed_t *cmd)
{
    /* Set arguments to zero */
    cmd->arg_integer = 0;
    cmd->arg_fractional = 0;
    cmd->lzeros = 0;
    cmd->arg_sign = false;
    cmd->string[0] = '\0';
    
    if(str == NULL)
        return;
    
    unsigned int cmdlen;
    if(cmd->suffix == SUF_VOID)
        cmdlen = strlen(prefix_literal(cmd->prefix));
    else
        cmdlen = strlen(suffix_literal(cmd->suffix));
    strncpy(cmd->string, str + cmdlen + 1, STRARG_LEN);
}

/*******************************************************
 * This function reads the received string from a 
 * buffer of type Com_Buffer_t, tokenizes it and returns 
 * the results as a struct of type Command_Parsed_t. The 
 * flags of read_buffer are not touched.
 *******************************************************/

static Command_Parsed_t parse_cmd(Com_Buffer_t *read_buffer)
{
    Command_Parsed_t cmd;
    
    /* Start parsing; tokenize string at ':'-character */
    char *token_1 = strtok(read_buffer->str, ":");
    char *token_2 = strtok(NULL, ":");
    
    /* Search for command literals in prefix and suffix */
    cmd.prefix = get_prefix(token_1);
    cmd.suffix = get_suffix(token_2);
    
    /* Determine if the argument is to be interpreted as a string or a number */
    bool arg_isstring;
    if(cmd.suffix == SUF_CALIBRATION_STRING)
        arg_isstring = true;
    else
        arg_isstring = false;
    
    /* Read and convert argument */
    if(token_2 != NULL)
    {
        /* The command consists of a prefix and a suffix */
        if(arg_isstring == false)
        {
            remove_space(token_2);
            cmd.query = is_query(token_2);
            cmd.unit = get_unit(token_2);
            extract_digits(token_2);
            convert_num_arg(token_2, &cmd);
        }
        else
        {
            remove_ctrl(token_2);
            cmd.query = is_query(token_2);
            cmd.unit = UNIT_VOID;
            convert_str_arg(token_2, &cmd);
        }
    }
    else
    {
        /* The command consists of a prefix only */
        if(arg_isstring == false)
        {
            remove_space(token_1);
            cmd.query = is_query(token_1);
            cmd.unit = get_unit(token_1);
            extract_digits(token_1);
            convert_num_arg(token_1, &cmd);
        }
        else
        {
            remove_ctrl(token_1);
            cmd.query = is_query(token_1);
            cmd.unit = UNIT_VOID;
            convert_str_arg(token_1, &cmd);
        }
    }
    return cmd;
}


/*******************************************************
 * These functions convert the argument in a struct of 
 * type Command_Parsed_t to various quantities like 
 * frequency in Hz, level in dBm, or a float.
 *******************************************************/

static int64_t pow10ull(unsigned int n)
{
    int powers[] = {1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000, 1000000000};
    if(n <= 9)
        return powers[n];
    else
        return pow(10, n);
}


static unsigned int numdigits(int64_t num)
{
    if(num < 0)
        return numdigits((num == LLONG_MIN) ? LLONG_MAX : -num);
    if(num < 10)
        return 1;
    return 1 + numdigits(num / 10);
}


static uint64_t cmdarg_to_frequency(const Command_Parsed_t *cmd)
{
    int exponent;
    unsigned int numfrac = numdigits(cmd->arg_fractional) + cmd->lzeros;
    uint64_t fractional;
    switch(cmd->unit)
    {
        case UNIT_GHZ:
            exponent = 9;
            break;
        case UNIT_MHZ:
            exponent = 6;
            break;
        case UNIT_KHZ:
            exponent = 3;
            break;
        case UNIT_HZ:
            exponent = 0;
            break;
        default:
            return 0;
    }
    if(numfrac > exponent)
    {
        fractional = cmd->arg_fractional / pow10ull(numfrac - exponent);
        numfrac = exponent;
    }
    else
        fractional = cmd->arg_fractional;
    return (cmd->arg_integer * pow10ull(exponent)) + (fractional * pow10ull(exponent - numfrac));
}


static float cmdarg_to_dbm(const Command_Parsed_t *cmd)
{
    const float system_impedance = 50;
    const float reference_power = 0.001;
    unsigned int numfrac = numdigits(cmd->arg_fractional) + cmd->lzeros;
    float arg_to_float = (float)(cmd->arg_integer) + ((float)(cmd->arg_fractional) / (float)pow10ull(numfrac));
    
    if(cmd->unit == UNIT_DBM)
    {
        if(cmd->arg_sign == true)
            arg_to_float *= -1;
        return arg_to_float;
    }
    else
    {
        int exponent;
        switch(cmd->unit)
        {
            case UNIT_V:
                exponent = 0;
                break;
            case UNIT_MV:
                exponent = -3;
                break;
            case UNIT_UV:
                exponent = -6;
                break;
            default:
                return 100;
        }
        float lvl_volt = arg_to_float * pow(10, exponent);
        float lvl_dbm = 10 * log10(pow(lvl_volt, 2) / (system_impedance * reference_power));
        return lvl_dbm;
    }
}


static float cmdarg_to_ms(const Command_Parsed_t *cmd)
{
    int exponent;
    unsigned int numfrac = numdigits(cmd->arg_fractional) + cmd->lzeros;
    float arg_to_float = (float)(cmd->arg_integer) + ((float)(cmd->arg_fractional) / pow10ull(numfrac));
    switch(cmd->unit)
    {
        case UNIT_S:
            exponent = 0;
            break;
        case UNIT_MS:
            exponent = -3;
            break;
        case UNIT_US:
            exponent = -6;
            break;
        default:
            return 0;
    }
    return arg_to_float * pow(10, exponent) * 1000;
}


static float cmdarg_to_float(const Command_Parsed_t *cmd)
{
    unsigned int numfrac = numdigits(cmd->arg_fractional) + cmd->lzeros;
    float arg_to_float = (float)(cmd->arg_integer) + ((float)(cmd->arg_fractional) / pow10ull(numfrac));
    if(cmd->arg_sign == true)
        arg_to_float *= -1;
    return arg_to_float;
}


/*******************************************************
 * These functions execute the commands handed over as 
 * a pointer. If any data is to be sent over the
 * USB CDC data link, it is written to write_buffer.
 *******************************************************/

static void execute_prefix_only(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    
    if(cmd->query == true)
    {
        /* Command is a query */
        switch(cmd->prefix)
        {
            case PRF_IDENTIFY:
                sprintf(write_buffer->str, "Signal Generator Mark III, Dr. Mario Hellmich, Salzgitter\r\n");
                return;
                
            case PRF_VERSION:
                sprintf(write_buffer->str, "Firmware Version %u.%u\r\n", fwversion.major, fwversion.minor);
                return;
                
            case PRF_ERROR:
                sprintf(write_buffer->str, "Error Code %u\r\n", error.errstatus);
                error.ARGUMENT_OUT_OF_RANGE = false;
                return;
            
            case PRF_GET_STATUS:
                print_errmsg();
                return;
                
            case PRF_RF_OUTPUT:
                sprintf(write_buffer->str, "RF Output %u\r\n", state.rfon);
                return;
                
            default:
                return;
        }
    }
    else
    {
        /* Command is not a query */
        switch(cmd->prefix)
        {
            case PRF_RESET:
                system_reset();
                return;
                
            case PRF_CLEAR_STATUS:
                error.errstatus = 0;
                return;
                
            case PRF_RF_OUTPUT:
                state.rfon = !!(cmd->arg_integer);
                synthesizer_update();
                return;
                
            case PRF_SAVE:
                if(mem_save(cmd->arg_integer) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                            
            case PRF_RECALL:
                if(mem_recall(cmd->arg_integer) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                
            default:
                return;
        }
    }
}


static void execute_suffix_cw(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    if(cmd->query == true)
    {
        /* Command is a query */
        switch(cmd->suffix)
        {
            case SUF_SWITCH_ON:
                sprintf(write_buffer->str, "CW %u\r\n", !state.sweep);
                return;
                
            case SUF_FREQUENCY:
                sprintf(write_buffer->str, "Frq %lu Hz\r\n", state.cw_freq);
                return;
                
            case SUF_LEVEL:
                sprintf(write_buffer->str, "Level %.2f dBm\r\n", state.level);
                return;
                
            default:
                return;
        }
    }
    else
    {
        /* Command is not a query */
        switch(cmd->suffix)
        {
            case SUF_SWITCH_ON:
                set_sweep_mode(false);
                return;
                
            case SUF_FREQUENCY:
                if(set_cw_freq_hz(cmdarg_to_frequency(cmd)) == EXIT_FAILURE)
                     error.ARGUMENT_OUT_OF_RANGE = true; 
                return;
                
            case SUF_LEVEL: ;
                if(set_lvl_db(cmdarg_to_dbm(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            default:
                return;
        }
    }
}


static void execute_suffix_sweep(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    if(cmd->query == true)
    {
        /* Command is a query */
        switch(cmd->suffix)
        {
            case SUF_SWITCH_ON:
                sprintf(write_buffer->str, "Sweep %u\r\n", state.sweep);
                return;
                
            case SUF_START_FREQUENCY:
                sprintf(write_buffer->str, "Start Frq %lu Hz\r\n", state.start_freq);
                return;
                
            case SUF_STOP_FREQUENCY:
                sprintf(write_buffer->str, "Stop Frq %lu Hz\r\n", state.stop_freq);
                return;
                
            case SUF_LEVEL:
                sprintf(write_buffer->str, "Level %.2f dBm\r\n", state.level);
                return;
                
            case SUF_DWELL_TIME:
                sprintf(write_buffer->str, "Dwell Time %.0f ms\r\n", state.dwell);
                return;
                
            case SUF_NPOINTS:
                sprintf(write_buffer->str, "N Points %u\r\n", state.npoints);
                return;
                
            case SUF_SINGLE:
                sprintf(write_buffer->str, "Single %u\r\n", state.single_sweep);
                return;
                
            default:
                return;
        }
    }
    else
    {
        /* Command is not a query */
        switch(cmd->suffix)
        {
            case SUF_SWITCH_ON:
                set_sweep_mode(true);
                return;
                
             case SUF_LEVEL:
                if(set_lvl_db(cmdarg_to_dbm(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            case SUF_START_FREQUENCY:
                if(set_start_frq_hz(cmdarg_to_frequency(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                
            case SUF_STOP_FREQUENCY:
                if(set_stop_frq_hz(cmdarg_to_frequency(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;                
                
            case SUF_DWELL_TIME:
                if(set_dwell_time_ms(cmdarg_to_ms(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            case SUF_NPOINTS:
                if(set_npoints(cmd->arg_integer) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                
            case SUF_SINGLE:
                set_sweep_single(!!(cmd->arg_integer));
                return;
                
            default:
                return;
        }
    }
}


static void execute_suffix_ref(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    if(cmd->query == true)
    {
        /* Command is a query */
        switch(cmd->suffix)
        {
            case SUF_REF_EXTERNAL:
                sprintf(write_buffer->str, "Ext Ref %u\r\n", state.extref);
                return;
                
            case SUF_FREQUENCY:
                sprintf(write_buffer->str, "Ref Frq %llu Hz\r\n", f_pfd / state.refmult);
                return;
                
            default:
                return;
        }
    }
    else
    {
        /* Command is not a query */
        switch(cmd->suffix)
        {
            case SUF_REF_EXTERNAL:
                set_extref(!!(cmd->arg_integer));
                return;
                
            case SUF_FREQUENCY:
                if(set_ref_frq_hz(cmdarg_to_frequency(cmd)) == EXIT_FAILURE)
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            default:
                return;
        }
    }
}


static void execute_suffix_cal(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    /* Handle switchover to calibration mode */
    if(cmd->suffix == SUF_SWITCH_ON)
    {
        if(cmd->query == true)
        {
            sprintf(write_buffer->str, "Cal %u\r\n", state.cal);
            return;
        }
        else if(cmd->arg_integer == password)
        {
            set_state_to_default(&state, &error);
            state.cw_freq = FREQ_MIN;
            state.cal = true;
            error.CAL_PROTECTED = false;
            synthesizer_update();
            return;
        }
        else
        {
            error.CAL_PROTECTED = true;
            state.cal = false;
            return;
        }
    }
    /* Check if in calibration mode, else return */
    if(state.cal == false)
        return;
    /* Start further command execution */
    if(cmd->query == true)
    {
        /* Command is a query */
        switch(cmd->suffix)
        {
            case SUF_LEVEL: ;
                float corr_val = get_lvl_corr_val(state.cw_freq, state.cal_att_stage);
                sprintf(write_buffer->str, "Level Corr %.4f at Att Stage %u\r\n", corr_val, state.cal_att_stage);
                return;
            
            case SUF_CAL_MODE:
                sprintf(write_buffer->str, "Calibration Mode %u\r\n", state.cal_mode);
                return;
            
            case SUF_ATT_STAGE:
                sprintf(write_buffer->str, "Att Stage %u\r\n", state.cal_att_stage);
                return;
                
            case SUF_SLOPE_CORR:
                sprintf(write_buffer->str, "Slope Corr %.2f\r\n", caldata_ram.slopecorr);
                return;
                
            case SUF_FREQUENCY:
                sprintf(write_buffer->str, "Frequency Corr %.2f Hz\r\n", caldata_ram.frqcorr);
                return;
                
            case SUF_CALIBRATION_STRING:
                sprintf(write_buffer->str, "Cal String: %s\r\n", caldata_ram.calstring);
                return;
                
            default:
                return;
        }
    }
    else
    {
        /* Command is not a query */
        switch(cmd->suffix)
        {
            case SUF_LEVEL:    ;
                float new_lvl_corr_val = cmdarg_to_float(cmd);
                if( (new_lvl_corr_val <= LVL_CORR_MAX) && (new_lvl_corr_val >= LVL_CORR_MIN) )
                {
                    set_lvl_corr_val(state.cw_freq, state.cal_att_stage, new_lvl_corr_val);
                    synthesizer_update();
                }
                else
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                
            case SUF_CAL_MODE:  ;
                int new_cal_mode = cmd->arg_integer;
                if( (new_cal_mode >= 0) && (new_cal_mode < numCalmode) )
                {
                    state.cal_mode = new_cal_mode;
                    synthesizer_update();
                }
                else
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            case SUF_ATT_STAGE: ;
                int new_att_stage = cmd->arg_integer;
                if( (new_att_stage >= 0) && (new_att_stage <= att_nstages) )
                {
                    state.cal_att_stage = new_att_stage;
                    synthesizer_update();
                }
                else
                    error.ARGUMENT_OUT_OF_RANGE;
                return;
                
            case SUF_SLOPE_CORR:    ;
                float new_slcorr = cmdarg_to_float(cmd);
                if( (new_slcorr <= SLOPE_CORR_MAX) && (new_slcorr >= SLOPE_CORR_MIN) )
                {
                    caldata_ram.slopecorr = new_slcorr;
                    synthesizer_update();
                }
                else
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
                
            case SUF_FREQUENCY:    ;
                float new_freqcorr = cmdarg_to_float(cmd);
                if( (new_freqcorr <= FREQ_CORR_MAX) && (new_freqcorr >= FREQ_CORR_MIN) )
                {
                    caldata_ram.frqcorr = new_freqcorr;
                    synthesizer_update();
                }
                else
                    error.ARGUMENT_OUT_OF_RANGE = true;
                return;
            
            case SUF_CALIBRATION_STRING:
                strncpy(caldata_ram.calstring, cmd->string, STRARG_LEN);
                return;
                
            case SUF_SAVE_CALDATA:
                save_cal_data();
                return;
            
            case SUF_DUMP_CALDATA:
                dump_cal_data();
                return;
                
            case SUF_ERASE_EEPROM:
                eeprom_erase();
                return;
                
            default:
                return;
        }
    }
}


static void execute_cmd(const Command_Parsed_t *cmd, Com_Buffer_t *write_buffer)
{
    /* Check if there is anything to execute, else return */
    if(cmd->prefix == PRF_VOID)
        return;
    
    /* Start executing commands. First check if there is only a prefix. */
    if(cmd->suffix == SUF_VOID)
    {
        /* Command consists of a prefix alone */
        execute_prefix_only(cmd, write_buffer);
    }
    else
    {
        /* Command consists of prefix and suffix. Start by reading the prefix. */
        switch(cmd->prefix)
        {
            case PRF_CW:
                execute_suffix_cw(cmd, write_buffer);
                break;
                
            case PRF_SWEEP:
                execute_suffix_sweep(cmd, write_buffer);
                break;
                
            case PRF_REFERENCE:
                execute_suffix_ref(cmd, write_buffer);
                break;
                
            case PRF_CAL:
                execute_suffix_cal(cmd, write_buffer);
                break;
                
            default:
                break;
        }
    }
}


/*******************************************************
 * This function is to be called periodically. It 
 * processes the command parser state machine.
 *******************************************************/

void CMD_Parser_Tasks(void)
{
    static Command_Parsed_t new_command;
    static CMD_Parser_State_t cmd_parser_state = CMD_PARSER_READ_NEW_DATA;
    
    /* Update local read buffer from queue */
    rx_buffer_update(&local_read_buffer);
    
    /* Command parser state machine */
    switch(cmd_parser_state)
    {
        case CMD_PARSER_READ_NEW_DATA:
            if(local_read_buffer.ready == true)                                 // Is new rx data available for parsing?
                cmd_parser_state = CMD_PARSER_PARSE_COMMAND;                    // Then next state is command parsing
            break;
        
        
        case CMD_PARSER_PARSE_COMMAND:
            new_command = parse_cmd(&local_read_buffer);                        // Parse newly received data
            if(new_command.prefix != PRF_VOID)                                  // Any command detected?
                cmd_parser_state = CMD_PARSER_EXECUTE_COMMAND;                  // Then next state is command execution
            else                                                                // No command was detected
                cmd_parser_state = CMD_PARSER_READ_NEW_DATA;                    // Then next state is read new data
            local_read_buffer.ready = false;
            break;
        
        
        case CMD_PARSER_EXECUTE_COMMAND:
            if(local_write_buffer.ready == false)                               // Is local write buffer available?
            {
                local_write_buffer.str[0] = '\0';                               // Clear string in buffer
                execute_cmd(&new_command, &local_write_buffer);                 // Execute command
                local_write_buffer.ready = true;                                // Set flag
                cmd_parser_state = CMD_PARSER_READ_NEW_DATA;                    // Next state is read new data
            }
            tx_buffer_update(&local_write_buffer);                              // Update local write buffer, send any tx data to queue
            break;
    }
}


/*******************************************************************************
 End of File
 */