/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app_eemem.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file contains control routines for the external EEPROM memory. 
 * 
 * *****************************************************************************/

/* Section: Includes ***********************************************************/

#include "app_eemem.h"


/* Section: Defines ************************************************************/
    
#define EEMEM_OFFSET    0xA0000000
#define EEMEM           __attribute__((section(".eeprom"), address(EEMEM_OFFSET), noload))


/* Section: Variables **********************************************************/

Calibration_Data_t caldata_ram;                                                 // RAM storage of calibration data

struct EEPROM_Storage_s {
    uint32_t checksum;
    uint8_t caldata[sizeof(caldata_ram)];
    Instrument_State_t instr_setup[NMEM];
};

static struct EEPROM_Storage_s EEMEM eemem;                                     // EEPROM storage of calibration and instrument setup data


/* Section: Function implementations *******************************************/

/*******************************************************
 * General EEPROM read and write functions.
 *******************************************************/

static inline bool eeprom_read(DRV_HANDLE handle, void *rxdata, uint32_t len, void *addr)
{
    return DRV_AT25_Read(handle, rxdata, len, (uintptr_t)(addr - EEMEM_OFFSET));
}


static inline bool eeprom_write(DRV_HANDLE handle, void *txdata, uint32_t len, void *addr)
{
    return DRV_AT25_Write(handle, txdata, len, (uintptr_t)(addr - EEMEM_OFFSET));
}


static inline void eeprom_busy_wait(DRV_HANDLE handle)
{
    while(DRV_AT25_TransferStatusGet(handle) == DRV_AT25_TRANSFER_STATUS_BUSY);
}


/* Calculate the CRC32 checksum. From mikrocontroller.net, credits: 
 * https://www.mikrocontroller.net/topic/225796?goto=new#2269908 
 * Do not use the hardware CRC calculator in the DSU of the
 * ATSAME54P20A for easier code and better portability. */

static const unsigned int gen_polynomial = 0xEDB88320;

static uint32_t crc32_chksum(const void *ptr, unsigned int len)
{
    uint32_t crc, mask;
    const uint8_t *uchar_ptr = ptr;
    crc = 0xFFFFFFFF;
    for(unsigned int i = 0; i < len; i++)
    {
        crc = crc ^ *(uchar_ptr + i);
        for(int j = 7; j >= 0; j--)
        {
            mask = -(crc & 1);
            crc = (crc >> 1)^(gen_polynomial & mask);
        }
    }
    return ~crc;
}


/*******************************************************
 * Load and save calibration data from and to SRAM.
 *******************************************************/

void load_cal_data(void)
{
    DRV_HANDLE drv_handle = DRV_AT25_Open(DRV_AT25_INDEX, DRV_IO_INTENT_READ | DRV_IO_INTENT_EXCLUSIVE);
    if(drv_handle == DRV_HANDLE_INVALID)
    {
        error.EEPROM_ERROR = true;
        return;
    }
        
    uint32_t checksum_eemem;
    
    /* Load calibration data from EEPROM */
    if(eeprom_read(drv_handle, &caldata_ram, sizeof(caldata_ram), &eemem.caldata) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    /* Load checksum from EEPROM */
    if(eeprom_read(drv_handle, &checksum_eemem, sizeof(checksum_eemem), &eemem.checksum) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    DRV_AT25_Close(drv_handle);
    
    /* Test checksum */    
    uint32_t checksum_calc = crc32_chksum(&caldata_ram, sizeof(caldata_ram));
    if(checksum_calc == checksum_eemem)
        error.EEPROM_CALDATA_CHKSUM = false;
    else
    {
        error.EEPROM_CALDATA_CHKSUM = true;
        memset(&caldata_ram, 0x00, sizeof(caldata_ram));
        strncpy(caldata_ram.calstring, "!! NO OR INVALID CAL DATA !!", STRARG_LEN);
    }
}


void save_cal_data(void)
{
    DRV_HANDLE drv_handle = DRV_AT25_Open(DRV_AT25_INDEX, DRV_IO_INTENT_WRITE | DRV_IO_INTENT_EXCLUSIVE);
    if(drv_handle == DRV_HANDLE_INVALID)
    {
        error.EEPROM_ERROR = true;
        return;
    }
    
    /* Write calibration data */
    if(eeprom_write(drv_handle, &caldata_ram, sizeof(caldata_ram), &eemem.caldata) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    /* Write checksum */
    uint32_t checksum = crc32_chksum(&caldata_ram, sizeof(caldata_ram));
    
    if(eeprom_write(drv_handle, &checksum, sizeof(checksum), &eemem.checksum) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    DRV_AT25_Close(drv_handle);
}


/*******************************************************
 * Instrument setup memory.
 *******************************************************/

static void read_setup_memory(unsigned int mem, Instrument_State_t *state_buf)
{
    if(mem > NMEM - 1)
        return;
    
    DRV_HANDLE drv_handle = DRV_AT25_Open(DRV_AT25_INDEX, DRV_IO_INTENT_READ | DRV_IO_INTENT_EXCLUSIVE);
    if(drv_handle == DRV_HANDLE_INVALID)
    {
        error.EEPROM_ERROR = true;
        return;
    }
    
    if(eeprom_read(drv_handle, &state_buf->cw_freq, sizeof(state_buf->cw_freq), &eemem.instr_setup[mem].cw_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->level, sizeof(state_buf->level), &eemem.instr_setup[mem].level) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->start_freq, sizeof(state_buf->start_freq), &eemem.instr_setup[mem].start_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->stop_freq, sizeof(state_buf->stop_freq), &eemem.instr_setup[mem].stop_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->dwell, sizeof(state_buf->dwell), &eemem.instr_setup[mem].dwell) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->npoints, sizeof(state_buf->npoints), &eemem.instr_setup[mem].npoints) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->refmult, sizeof(state_buf->refmult), &eemem.instr_setup[mem].refmult) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
        
    if(eeprom_read(drv_handle, &state_buf->sweep, sizeof(state_buf->sweep), &eemem.instr_setup[mem].sweep) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->single_sweep, sizeof(state_buf->single_sweep), &eemem.instr_setup[mem].single_sweep) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->rfon, sizeof(state_buf->rfon), &eemem.instr_setup[mem].rfon) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_read(drv_handle, &state_buf->extref, sizeof(state_buf->extref), &eemem.instr_setup[mem].extref) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    DRV_AT25_Close(drv_handle);
}


static void write_setup_memory(unsigned int mem, Instrument_State_t *state_buf)
{
    if(mem > NMEM -1)
        return;
    
    DRV_HANDLE drv_handle = DRV_AT25_Open(DRV_AT25_INDEX, DRV_IO_INTENT_WRITE | DRV_IO_INTENT_EXCLUSIVE);
    if(drv_handle == DRV_HANDLE_INVALID)
    {
        error.EEPROM_ERROR = true;
        return;
    }
    
    if(eeprom_write(drv_handle, &state_buf->cw_freq, sizeof(state_buf->cw_freq), &eemem.instr_setup[mem].cw_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->level, sizeof(state_buf->level), &eemem.instr_setup[mem].level) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->start_freq, sizeof(state_buf->start_freq), &eemem.instr_setup[mem].start_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->stop_freq, sizeof(state_buf->stop_freq), &eemem.instr_setup[mem].stop_freq) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->dwell, sizeof(state_buf->dwell), &eemem.instr_setup[mem].dwell) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->npoints, sizeof(state_buf->npoints), &eemem.instr_setup[mem].npoints) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->refmult, sizeof(state_buf->refmult), &eemem.instr_setup[mem].refmult) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
        
    if(eeprom_write(drv_handle, &state_buf->sweep, sizeof(state_buf->sweep), &eemem.instr_setup[mem].sweep) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->single_sweep, sizeof(state_buf->single_sweep), &eemem.instr_setup[mem].single_sweep) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->rfon, sizeof(state_buf->rfon), &eemem.instr_setup[mem].rfon) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    if(eeprom_write(drv_handle, &state_buf->extref, sizeof(state_buf->extref), &eemem.instr_setup[mem].extref) == false)
        error.EEPROM_ERROR = true;
    else
        eeprom_busy_wait(drv_handle);
    
    DRV_AT25_Close(drv_handle);
}


static bool state_isequal(const Instrument_State_t *a, const Instrument_State_t *b)
{
    if(a->cw_freq != b->cw_freq)    
        return false;
    if(a->level != b->level)
        return false;
    if(a->start_freq != b->start_freq)
        return false;
    if(a->stop_freq != b->stop_freq)
        return false;
    if(a->dwell != b->dwell)
        return false;
    if(a->npoints != b->npoints)
        return false;
    if(a->refmult != b->refmult)
        return false;
    if(a->sweep != b->sweep)
        return false;
    if(a->single_sweep != b->single_sweep)
        return false;
    if(a->rfon != b->rfon)
        return false;    
    if(a->extref != b->extref)
        return false;
    /* No comparison for calmode, cal_mode and cal_att_stage, since it is not stored in EEPROM */
    return true;
}


static bool state_isvalid(const Instrument_State_t *a)
{
    if(a->cw_freq < FREQ_MIN || a->cw_freq > FREQ_MAX)
        return false;
    if(a->level < LVL_MIN || a->level > LVL_MAX)
        return false;
    if(a->start_freq < FREQ_MIN || a->start_freq > FREQ_MAX)
        return false;
    if(a->stop_freq < FREQ_MIN || a->stop_freq > FREQ_MAX)
        return false;
    if(a->start_freq > a->stop_freq)
        return false;
    if(a->dwell < DWELL_MIN || a->dwell > DWELL_MAX)
        return false;
    if(a->npoints < NPOINTS_MIN || a->npoints > NPOINTS_MAX)
        return false;
    if(a->refmult < REFMULT_MIN || a->refmult > REFMULT_MAX)
        return false;
    return true;
}


/*******************************************************
 * Externally accessible functions to set and recall
 * instrument states from setup memory.
 *******************************************************/

bool mem_recall(unsigned int mem)
{
    if(mem > NMEM - 1)
        return EXIT_FAILURE;
    
    Instrument_State_t state_buf;
    read_setup_memory(mem, &state_buf);
    state_buf.cal = false;
    state_buf.cal_att_stage = 0;
    state_buf.cal_mode = 0;
	if(state_isvalid(&state_buf) == true)
        state = state_buf;
    else
        set_state_to_default(&state, NULL);
    synthesizer_update();
    
    return EXIT_SUCCESS;
}


bool mem_save(unsigned int mem)
{
    if(mem > NMEM - 1)
        return EXIT_FAILURE;
    
    taskENTER_CRITICAL();                                                       // Disable interrupts and context switches
        Instrument_State_t state_current = state;
    taskEXIT_CRITICAL();                                                        // Enable interrupts and context switches
    Instrument_State_t state_in_mem;
    read_setup_memory(mem, &state_in_mem);
    if(state_isequal(&state_current, &state_in_mem) == false)
        write_setup_memory(mem, &state_current);
    
    return EXIT_SUCCESS;
}


/*******************************************************
 * EEPROM service functions.
 *******************************************************/

void eeprom_erase(void)
{
    DRV_HANDLE drv_handle = DRV_AT25_Open(DRV_AT25_INDEX, DRV_IO_INTENT_READWRITE | DRV_IO_INTENT_EXCLUSIVE);
    if(drv_handle == DRV_HANDLE_INVALID)
    {
        error.EEPROM_ERROR = true;
        return;
    }
    
    uint8_t write_buffer[DRV_AT25_EEPROM_PAGE_SIZE] = {[0 ... DRV_AT25_EEPROM_PAGE_SIZE - 1] = 0xFF};
    const int max_addr = DRV_AT25_EEPROM_FLASH_SIZE;
    
    for(int addr = 0; addr < max_addr; addr += DRV_AT25_EEPROM_PAGE_SIZE)
    {
        if(DRV_AT25_PageWrite(drv_handle, write_buffer, addr) == false)
            error.EEPROM_ERROR = true;
        else
            eeprom_busy_wait(drv_handle);
    }
    DRV_AT25_Close(drv_handle);
}

/*******************************************************************************
 End of File
 */