/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app_hwcontrol.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file contains control routines for the synthesizer hardware. 
 * In particular, the DDS synthesizer, the main PLL synthesizer, reference 
 * PLL and DAC.
 * 
 * *****************************************************************************/

/*Section: Includes ************************************************************/

#include "app_hwcontrol.h"


/*Section: Function implementations ********************************************/

/*******************************************************
 * DDS synthesizer
 *******************************************************/

static void dds_write_reg(uint8_t instruction, const uint8_t *buffer, unsigned int len)
{
    spi_send_reg(&instruction, 1, SPI_ADDR_DDS);
    spi_send_reg(buffer, len, SPI_ADDR_DDS);
}


static void dds_assert_update(void)
{
    spi_send_reg(0, 1, SPI_ADDR_DDSUPDATE);
    blocking_delay(4);
    clr_chip_select();
}


void dds_init(void)
{
    DDS_Update_Clear();
    DDS_Reset_Set();
    vTaskDelay(10 / portTICK_PERIOD_MS);
    DDS_Reset_Clear();
    vTaskDelay(20 / portTICK_PERIOD_MS);
    Convert32to8_t register1, register2;
    register1.i32 = 0;
    register2.i32 = 0;
    register1.i32 |= (1<<1)  |                          // Disable SYNC_CLK
                     (1<<3)  |                          // External power down mode; full power down commanded by pin
                     (1<<9)  |                          // SDIO configured as input only
                     (1<<13);                           // Auto-clear phase accumulator
    register2.i32 |= (1<<1)  | (1<<0) |                 // Set charge pump current to 150 A
                     (1<<2)  |                          // Set VCO range to 250 ... 400 MHz
                     (0x04 << 3);                       // Set reference clock multiplier to 4, --> SysCLK = 400 MHz
                     
    dds_write_reg(0x00, register1.byte, 4);
    dds_write_reg(0x01, register2.byte, 3);
    clr_chip_select();
    dds_assert_update();
}


void dds_write_tunewd(uint32_t dds_tunewd)
{
    Convert32to8_t buf;
    buf.i32 = dds_tunewd;    
    dds_write_reg(0x04, buf.byte, 4);                                           // Write tuning word to register address 0x04
    clr_chip_select();
    dds_assert_update();                                                        // Assert DDS Update line to transfer FIFO register contents
}


/*******************************************************
 * Reference PLL
 *******************************************************/

static void refpll_write_reg(const uint32_t *reg)
{
    Convert32to8_t buffer;
    buffer.i32 = *reg;
    spi_send_reg(buffer.byte, 3, SPI_ADDR_REFPLL);
    clr_chip_select();
}


void refpll_write_mult(uint32_t mult)
{
    uint32_t reg = 0;
    mult = mult & 0x00001FFF;                           // Truncate to 13 bit
    reg = (1UL<<0) | (0UL<<1)                           // Address N-counter latch
          | (mult << 8)                                 // Write N-multiplier
          | (0UL<<21);                                  // Charge pump setting 1 is permanently used
    refpll_write_reg(&reg);
}


void refpll_set_ext_ref(bool extref)
{
    if(extref == false)
    {
        Int_Sel_Set();
        Ext_Sel_Clear();
    }
    else
    {
        Int_Sel_Clear();
        Ext_Sel_Set();
    }
}


void refpll_init(void)
{
    uint32_t ref_counter_latch = 0, n_counter_latch = 0, funct_latch = 0, init_latch = 0;
    
    ref_counter_latch = (0UL<<0)  | (0UL<<1)      |     // Address reference counter latch
                        ( 1 & 0x3FFF <<2 )  |           // Set reference divider to 1
                        (0UL<<16) | (0UL<<17)     |     // 2.9ns anti-backslash pulse width
                        (0UL<<20);                      // Fast lock detect
    
    n_counter_latch   = (1UL<<0)  | (0UL<<1)      |     // Address n counter latch
                        (10 << 8)             |         // Set n counter to 10 (100 MHz VCO frequency)
                        (0UL<<21);                      // Charge pump current setting 1 is permanently used
    
    funct_latch       = (0UL<<0)  | (1UL<<1)      |     // Address function latch
                        (0UL<<2)  |                     // Counter normal operation
                        (0UL<<21) | (0UL<<3)      |     // No power-down, normal operation
                        (0UL<<6)|(0UL<<5)|(1UL<<4)|     // MUXOUT digital lock detect
                        (1UL<<7)  |                     // PFD polarity positive
                        (0UL<<8)  |                     // CP not in three state
                        (0UL<<9)  | (0UL<<10)     |     // Fast lock disabled
                        (0x0007 << 11)            |     // Charge pump current setting 1 (5 mA)
                        (0x0007 << 14);                 // Charge pump current setting 2 (5 mA)
    
    init_latch        = (0UL<<1)  | (1UL<<1)      |     // Address initialization latch
                        (0UL<<2)  |                     // Counter normal operation
                        (0UL<<21) | (0UL<<3)      |     // No power-down, normal operation
                        (0UL<<6)|(0UL<<5)|(1UL<<4)|     // MUXOUT digital lock detect
                        (1UL<<7)  |                     // PFD polarity positive
                        (0UL<<8)  |                     // CP not in three state
                        (0UL<<9)  | (0UL<<10)     |     // Fast lock disabled
                        (0x0007 << 11)            |     // Charge pump current setting 1 (5 mA)
                        (0x0007 << 14);                 // Charge pump current setting 2 (5 mA)
    
    refpll_write_reg(&init_latch);
    refpll_write_reg(&funct_latch);
    refpll_write_reg(&ref_counter_latch);
    refpll_write_reg(&n_counter_latch);
}


/*******************************************************
 * Main PLL
 *******************************************************/

static void mpll_write_reg(const uint32_t *reg)
{
    Convert32to8_t buffer;
    buffer.i32 = *reg;
    spi_send_reg(buffer.byte, 3, SPI_ADDR_MPLL);
    clr_chip_select();
}


void mpll_init(void)
{
    uint32_t  reg0,  reg1,  reg2,  reg4,  reg7,  reg8,
              reg9, reg10, reg11, reg12, reg13, reg14,
             reg19, reg20, reg22, reg23, reg24, reg25,
             reg28, reg29, reg30, reg31, reg32, reg33,
             reg34, reg35, reg36, reg37, reg38, reg39,
             reg40, reg41, reg42, reg43, reg44, reg45,
             reg46, reg47, reg48, reg59, reg61, reg62,
             reg64, reg68, reg69, reg70;
    
    reg0  = (0UL<<16)  |                                // Address register 0
            (1UL<<1)   |                                // Perform soft reset
            (1UL<<9);                                   // Default
    mpll_write_reg(&reg0);                              // Write register 0 to perform a soft reset
    vTaskDelay(100 / portTICK_PERIOD_MS);               // Wait 100 ms
    
    reg0  = (0UL<<16)  |                                // Address register 0
            (0UL<<0)   |                                // Device power up
            (1UL<<2)   |                                // MUX out pin lock detect
            (1UL<<3)   |                                // Perform VCO calibration
            (1UL<<4)   |                                // Amplitude calibration enable
            (0UL<<5)   |                                // FCAL_HPFD_ADJ, used when f_pfd is low, set for f_pfd > 10 MHz
            (2UL<<7)   |                                // FCAL_HPFD_ADJ, used when f_pfd is high, set to 150 MHz < f_pfd <= 200 MHz
            (1UL<<9)   |                                // Default
            (1UL<<13);                                  // Lock detect enable
    
    reg1  = (1UL<<16)  |                                // Address register 1
            (1UL<<0)   |                                // Set CAL_CLK_DIV to 2^1 = 2, i.e., cal clock is approx. 100 MHz
            (1UL<<3)   |                                // Default
            (1UL<<11);                                  // Default
    
    reg2  = 0x020500;                                   // Register as per TICSpro
    
    reg4  = (4UL<<16)  |                                // Address register 4
            ( (25 & 0xFF) << 8 ) |                      // Set VCO amplitude calibration delay (ACAL_CMP_DLY) to 25, then 2 * ACAL_CMP_DLY * [cal clock period] = 500 ns, whis is greater than the required minimum)
            (1UL<<0)   |                                // Default
            (1UL<<1)   |                                // Default
            (1UL<<6);                                   // Default
    
    reg7  = 0x0728B2;                                   // Register as per TICSpro
    
    reg8  = (8UL<<16)  |                                // Address register 8
            (1UL<<2)   |                                // Default
            (1UL<<7)   |                                // Default
            (0UL<<10)  |                                // Override VCO capacitor bank selection disable
            (1UL<<12)  |                                // Default
            (0UL<<13);                                  // Override VCO amplitude disable
    
    reg9  = (9UL<<16)  |                                // Address register 9
            (1UL<<1)   |                                // Default
            (1UL<<8)   |                                // Default
            (1UL<<9)   |                                // Enable reference path 
            (0UL<<11);                                  // Disable reference path doubler
    
    reg10 = (10UL<<16) |                                // Address register 10
            (1UL<<3)   |                                // Default
            (1UL<<4)   |                                // Default
            (1UL<<6)   |                                // Default
            ( (1 & 0x1F) << 7 ) |                       // Set reference multiplier to 1 (bypass))
            (1UL<<12);                                  // Default
    
    reg11 = (11UL<<16) |                                // Address register 11
            (1UL<<3)   |                                // Default
            ( (1 & 0xFF) << 4 );                        // Set R-divider after multiplier and before PFD to 1
    
    reg12 = (12UL<<16) |                                // Address register 12
            ( (1UL & 0x0FFF) ) |                        // Set R-divider after OSC-in doubler and before multiplier to 1
            (1UL<<12)  |                                // Default
            (1UL<<13)  |                                // Default
            (1UL<<14);                                  // Default
    
    reg13 = (13UL<<16) |                                // Address register 13
            (1UL<<14)  |                                // Enable charge pump
            (0UL<<11);                                  // Select dual CP mode
    
    reg14 = (14UL<<16) |                                // Address register 14
            (3UL<<0)   |                                // Set charge pump gain multiplier to times 2.5
            (0x1F<<2)  |                                // Set charge pump up current to maximum
            (0x1F<<7);                                  // Set charge pump down current to maximum
    
    reg19 = 0x130965;                                   // Register as per TICSpro
    reg20 = 0x14012C;                                   // Register as per TICSpro
    reg22 = 0x162300;                                   // Register as per TICSpro
    reg23 = 0x178842;                                   // Register as per TICSpro
    reg24 = 0x180509;                                   // Register as per TICSpro
    reg25 = 0x190000;                                   // Register as per TICSpro
    reg28 = 0x1C2924;                                   // Register as per TICSpro
    reg29 = 0x1D0084;                                   // Register as per TICSpro
    
    reg30 = (30UL<<16) |                                // Address register 30
            (1UL<<2)   |                                // Default
            (1UL<<4)   |                                // Default
            (1UL<<5)   |                                // Default
            (0UL<<10);                                  // Select MASH dithering off for integer-N mode
    
    reg31 = (31UL<<16) |                                // Address register 31
            (1UL<<0)   |                                // Default
            (0UL<<7)   |                                // Power up buffer between VCO and channel divider
            (0UL<<9)   |                                // Power up buffer between VCO and output A
            (1UL<<10);                                  // Power down buffer between VCO and output B
    
    reg32 = 0x20210A;                                   // Register as per TICSpro
    reg33 = 0x212A0A;                                   // Register as per TICSpro
    
    reg34 = (34UL<<16) |                                // Address register 31
            (1UL<<1)   |                                // Default
            (1UL<<3)   |                                // Default
            (1UL<<5)   |                                // Enable channel divider
            (1UL<<6)   |                                // Default
            (1UL<<7)   |                                // Default
            (1UL<<8)   |                                // Default
            (1UL<<9)   |                                // Default
            (1UL<<14)  |                                // Default
            (1UL<<15);                                  // Default
    
    reg35 = 0x23029B;                                   // Register as per TICSpro, will be overwritten later
    reg36 = 0x240C22;                                   // Register as per TICSpro, will be overwritten later
    
    reg37 = (37UL<<16) |                                // Address register 37
            (0UL<<12)  |                                // Set N divider prescaler, prescaler = 0: divide-by-2, prescaler = 1: divide-by-4
            (1UL<<14);                                  // Default
    
    reg38 = 0x260014;                                   // Register as per TICSpro, will be overwritten later
    
    reg39 = (39UL<<16) |                                // Address register 39
            (1UL<<2)   |                                // Default
            ( (1 & 0x3F) << 8 )  |                      // Set number of VCO cycles delay used by the PFD for compensation in a lock state to 4
            (1UL<<15);                                  // Default
    
    reg40 = (40UL<<16) |                                // Address register 40
            ( 0 & 0xFFFF );                             // Denominator of PLL divider fraction, the 16 MSB
    
    reg41 = (41UL<<16) |                                // Address register 41
            ( 1 & 0xFFFF );                             // Denominator of PLL divider fraction, the 16 LSB
    
    reg42 = 0x2A0000;                                   // Register as per TICSpro
    reg43 = 0x2B0000;                                   // Register as per TICSpro   
    
    reg44 = (44UL<<16) |                                // Address register 44
            ( 0 & 0xFFFF );                             // Numerator of PLL divider fraction, the 16 MSB
    
    reg45 = (45UL<<16) |                                // Address register 45
            ( 0 & 0xFFFF );                             // Numerator of PLL divider fraction, the 16 LSB
    
    reg46 = (46UL<<16) |                                // Address register 46
            (0UL<<0)   | (0UL<<1) | (0UL<<2) |          // Set MASH order to 0
            (1UL<<5)   |                                // Default
            (0UL<<6)   |                                // Power up output A
            (1UL<<7)   |                                // Power down output B
            ( (63 & 0x3F) << 8 );                       // Set output A power to 63 (min: 0, max: 63)
    
    reg47 = 0x2F00CF;                                   // Register as per TICSpro, will be overwritten later
    reg48 = 0x3003FC;                                   // Register as per TICSpro
    
    reg59 = (59UL<<16) |                                // Address register 59
            (0UL<<5);                                   // Select low current drive on MUXOUT pin
    
    reg61 = (61UL<<16) |                                // Address register 61
            (0UL<<0);                                   // Lock detect type, set to digital lock detect
    
    reg62 = 0x3E0000;                                   // Register as per TICSpro
    
    reg64 = (64UL<<16) |                                // Address register 64
            ( (7 & 0x0F) << 0 )  |                      // FJUMP_SIZE, increment size for jump during FCAL, used if FCAL_FAST = 1
            (1UL<<4)   |                                // Default
            ( (3 & 0x07) << 5 )  |                      // AJUMP_SIZE, increment size for jump during ACAL, used if ACAL_FAST = 1
            (1UL<<8)   |                                // FCAL_FAST enable
            (1UL<<9);                                   // ACAL_FAST enable
            
    reg68 = 0x440089;                                   // Register as per TICSpro
    reg69 = 0x450000;                                   // Register as per TICSpro
    reg70 = 0x460000;                                   // Register as per TICSpro
    
    mpll_write_reg(&reg1);                              // Write registers to chip
    mpll_write_reg(&reg2);
    mpll_write_reg(&reg4);
    mpll_write_reg(&reg7);
    mpll_write_reg(&reg8);
    mpll_write_reg(&reg9);
    mpll_write_reg(&reg10);
    mpll_write_reg(&reg11);
    mpll_write_reg(&reg12);
    mpll_write_reg(&reg13);
    mpll_write_reg(&reg14);
    mpll_write_reg(&reg19);
    mpll_write_reg(&reg20);
    mpll_write_reg(&reg22);
    mpll_write_reg(&reg23);
    mpll_write_reg(&reg24);
    mpll_write_reg(&reg25);
    mpll_write_reg(&reg28);
    mpll_write_reg(&reg29);
    mpll_write_reg(&reg30);
    mpll_write_reg(&reg31);
    mpll_write_reg(&reg32);
    mpll_write_reg(&reg33);
    mpll_write_reg(&reg34);
    mpll_write_reg(&reg35);
    mpll_write_reg(&reg36);
    mpll_write_reg(&reg37);
    mpll_write_reg(&reg38);
    mpll_write_reg(&reg39);
    mpll_write_reg(&reg40);
    mpll_write_reg(&reg41);
    mpll_write_reg(&reg42);
    mpll_write_reg(&reg43);
    mpll_write_reg(&reg44);
    mpll_write_reg(&reg45);
    mpll_write_reg(&reg46);
    mpll_write_reg(&reg47);
    mpll_write_reg(&reg48);
    mpll_write_reg(&reg59);
    mpll_write_reg(&reg61);
    mpll_write_reg(&reg62);
    mpll_write_reg(&reg64);
    mpll_write_reg(&reg68);
    mpll_write_reg(&reg69);
    mpll_write_reg(&reg70);
    
    mpll_write_reg(&reg0);                              // Write register 0 last to start amplitude and frequency calibration
}


void mpll_write(const struct mpll_s *params)
{
    uint32_t reg0, reg35, reg36, reg38, reg46, reg47;
    
    reg0  = (0UL<<16)  |                                // Address register 0
            (0UL<<0)   |                                // Device power up
            (1UL<<2)   |                                // MUX out pin lock detect
            (1UL<<3)   |                                // Perform VCO calibration
            (1UL<<4)   |                                // Amplitude calibration enable
            (0UL<<5)   |                                // FCAL_HPFD_ADJ, used when f_pfd is low, set for f_pfd > 10 MHz
            (2UL<<7)   |                                // FCAL_HPFD_ADJ, used when f_pfd is high, set to 150 MHz < f_pfd <= 200 MHz
            (1UL<<9)   |                                // Default
            (1UL<<13);                                  // Lock detect enable
    
    reg35 = (35UL<<16) |                                // Address register 35
            (1UL<<0)   |                                // Default
            (1UL<<1)   |                                // Channel divider segment 1 enable
            ( !!(params->chdiv_seg1) << 2 ) |           // Channel divider segment 1 division: chdiv_seg1 = 0: div by 2, chdiv_seg1 = 1: div by 3
            (1UL<<3)   |                                // Default
            (1UL<<4)   |                                // Default
            (1UL<<7)   |                                // Channel divider segment 2 enable
            (1UL<<8)   |                                // Channel divider segment 3 enable
            ( (params->chdiv_seg2 & 0x0F) << 9 );       // Channel divider segment 2 division
    
    reg36 = (36UL<<16) |                                // Address register 36
            ( (params->chdiv_seg3 & 0x0F) << 0 )    |   // Channel divider segment 3 division
            ( (params->chdiv_seg_sel & 0x07) << 4 ) |   // Channel divider multiplexer
            (1UL<<10);                                  // Enable buffer between channel divider and output A
    
    reg38 = (38UL<<16) |                                // Address register 38
            ( (params->pll_int & 0x0FFF) << 1 );        // PLL N divider, 12 bit
    
    reg46 = (46UL<<16) |                                // Address register 46
            (0UL<<0)   | (0UL<<1) | (0UL<<2) |          // Set MASH order to 0
            (1UL<<5)   |                                // Default
            ( (!(params->rfon) & 0x01) << 6 )|          // Select output A power up or down
            (1UL<<7)   |                                // Power down output B
            ( (params->drive_lvl & 0x3F) << 8 );        // Set output A power (min: 0, max: 63)
    
    reg47 = (47UL<<16) |                                // Address register 47
            ( !!(params->outa_mux) << 11 ) |            // Route signal to output from VCO or channel divider; outa_mux = 0: from channel divider, outa_mux = 1: from VCO
            (1UL<<6)   |                                // Default
            (1UL<<7);                                   // Default
    
    mpll_write_reg(&reg35);                             // Write registers to chip
    mpll_write_reg(&reg36);
    mpll_write_reg(&reg38);
    mpll_write_reg(&reg46);
    mpll_write_reg(&reg47);
    mpll_write_reg(&reg0);
}


/*******************************************************
 * Harmonic Filter Control
 *******************************************************/
 
#define FSW_INP_bm      (PORT_PB10 | PORT_PB11 | PORT_PB12)
#define FSW_OUT_bm      (PORT_PB13 | PORT_PB14 | PORT_PB15)

void hfilter_write(unsigned int hfiltnumber)
{
    if(hfiltnumber > 8)
        return;
    PORT_REGS->GROUP[1].PORT_OUTCLR = FSW_INP_bm;
    PORT_REGS->GROUP[1].PORT_OUTCLR = FSW_OUT_bm;
    if(hfiltnumber <= 1)
    {
        PORT_REGS->GROUP[1].PORT_OUTSET = FSW_INP_bm & ( 7 << (ffs(FSW_INP_bm) - 1) );        // Select output 8 on input mux
        PORT_REGS->GROUP[1].PORT_OUTSET = FSW_OUT_bm & ( 0 << (ffs(FSW_OUT_bm) - 1) );        // Select input 0 on output mux
        if(hfiltnumber == 0)
            nBand1_Sel_Clear();
        else
            nBand1_Sel_Set();
    }
    else
    {
        unsigned int val = hfiltnumber - 1;                                                   // Decrement hfiltnumber by 1 since filters 2 ... 8 are directly connected to the multiplexers
        PORT_REGS->GROUP[1].PORT_OUTSET = FSW_INP_bm & ( (7-val) << (ffs(FSW_INP_bm) - 1) );  // Select output on input mux
        PORT_REGS->GROUP[1].PORT_OUTSET = FSW_OUT_bm & ( val     << (ffs(FSW_OUT_bm) - 1) );  // Select input on output mux
    }
}


/*******************************************************
 * Attenuator Control
 *******************************************************/

#define ATT_CTRL_bm     (PORT_PB17 | PORT_PB18 | PORT_PB19 | PORT_PB20)

void atten_write(unsigned int val)
{
    if(val > 15)
        return;
    Att_LE_Clear();
    PORT_REGS->GROUP[1].PORT_OUTCLR = ATT_CTRL_bm;
    PORT_REGS->GROUP[1].PORT_OUTSET = ATT_CTRL_bm & (~val << (ffs(ATT_CTRL_bm) - 1));
    blocking_delay(6);
    Att_LE_Set();
    blocking_delay(6);
    Att_LE_Clear();
}


/*******************************************************
 * DAC Control
 *******************************************************/

static void dac_write_reg(const uint16_t *reg)
{
    Convert16to8_t buffer;
    buffer.i16 = *reg;
    spi_send_reg(buffer.byte, 2, SPI_ADDR_DAC);
    clr_chip_select();
}


void dac_write(uint16_t val)
{
    uint16_t reg = val & 0x0FFF;                        // Truncate to 12 bit
    reg |= (1<<14) | (1<<13) | (1<<12);                 // Set bit 14: input buffer enable, bit 13: output gain times 1, bit 12: no power down
    dac_write_reg(&reg);
}


/*******************************************************
 * Signal routing
 *******************************************************/

void bandselect_pllout(bool val)
{
    if(val == true)
        HF_Sel_Set();
    else
        HF_Sel_Clear();
}


/*******************************************************
 * Synthesizer power down
 *******************************************************/

void synthesizer_power_down(bool val)
{
    if(val == true)
    {
        PDN_Set();
        nPDN_Clear();
    }
    else
    {
        PDN_Clear();
        nPDN_Set();
    }
}


/*******************************************************************************
 End of File
 */