/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: instr_state.h
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This header file provides global variable definitions for the signal
 * generator state.
 * 
 * *****************************************************************************/

#ifndef APP_INSTR_STATE_H
#define	APP_INSTR_STATE_H


/*Section: Includes ************************************************************/

#include "definitions.h"


#ifdef	__cplusplus
extern "C" {
#endif

/*Section: Type definitions ****************************************************/

typedef enum {
    /* Normal calibration mode */
    CALMODE_NORMAL = 0,  
    /* Slope calibration mode, output level high */
    CALMODE_SLOPE_HIGH = 1,
    /* Slope calibration mode, output level low */
    CALMODE_SLOPE_LOW = 2,
    numCalmode,
} Cal_Mode_t;


typedef union {
    struct {
        bool ALC_UNLEVELED          : 1;    
        bool REFPLL_UNLK            : 1;
        bool MAINPLL_UNLK           : 1;
        bool ARGUMENT_OUT_OF_RANGE  : 1;
        bool CAL_PROTECTED          : 1;
        bool EEPROM_ERROR           : 1;
        bool EEPROM_CALDATA_CHKSUM  : 1;
    };
    unsigned int errstatus;
} Instrument_Error_t;


typedef struct {
    struct {
        uint32_t   cw_freq;           // Unit: Hz
        uint32_t   start_freq;        // Unit: Hz
        uint32_t   stop_freq;         // Unit: Hz
        float      dwell;             // Unit: ms
        uint16_t   npoints;           // Dimensionless
        uint32_t   refmult;           // Dimensionless
        float      level;             // Unit: dBm
        Cal_Mode_t cal_mode;          // Dimensionless
        uint8_t    cal_att_stage;     // Dimensionless
    };
    bool sweep;
    bool single_sweep;
    bool rfon;
    bool extref;
    bool cal;
} Instrument_State_t;


typedef struct {
    unsigned int major;
    unsigned int minor;
} Version_t;


/*Section: Variables ***********************************************************/
 
extern volatile Instrument_State_t state;       // Instrument state
extern volatile Instrument_Error_t error;       // Instrument error
extern const Version_t fwversion;               // Firmware version
extern const unsigned int password;             // Calibration protection password

extern const uint64_t f_pfd;                    // Main PLL nominal PFD frequency
extern const uint64_t f_int_ref;                // Internal reference frequency
extern const uint64_t f_dds_ref;                // DDS reference frequency
extern const uint64_t f_dds_sysclk;             // DDS system clock


/*Section: Defines *************************************************************/

#define FREQ_MAX                4200000000UL                                    // Maximum frequency (Hz)
#define FREQ_MIN                1000000UL                                       // Minimum frequency (Hz)
#define FREQ_BAND_CHANGE        20000000UL                                      // Frequency for switching between main PLL and DDS
#define REFMULT_MAX             1000                                            // Maximum reference multiplier
#define REFMULT_MIN             10                                              // Minimum reference multiplier
#define DWELL_MAX               100000                                          // Maximum dwell time (ms)
#define DWELL_MIN               1                                               // Minimum dwell time (ms)
#define NPOINTS_MAX             100001                                          // Maximum number of sweep points
#define NPOINTS_MIN             2                                               // Minimum number of sweep points
#define LVL_MAX                 10                                              // Maximum level (dBm)
#define LVL_MIN                 -45                                             // Minimum level (dBm)
#define NMEM                    100                                             // Number of instrument setup memories
#define LVL_CORR_MAX            50                                              // Maximum level correction value (dB)
#define LVL_CORR_MIN            -50                                             // Minimum level correction value (dB)
#define SLOPE_CORR_MAX          300                                             // Maximum detector slope correction value
#define SLOPE_CORR_MIN          -300                                            // Minimum detector slope correction value
#define SLOPE_CORR_LVL_DIFF     5                                               // Level difference between high and low output level for detector slope calibration
#define FREQ_CORR_MAX           1000                                            // Maximum time base correction value (Hz)
#define FREQ_CORR_MIN           -1000                                           // Minimum time base correction value (Hz)
#define FREQ_CORR_TRANS         100000000UL                                     // Transition frequency (Hz) between narrow and wide level calibration points spacing
#define FREQ_CORR_SPACING_LO    5000000UL                                       // Spacing between level calibration points (in Hz) below FREQ_CORR_TRANS
#define FREQ_CORR_SPACING_HI    100000000UL                                     // Spacing between level calibration points (in Hz) above FREQ_CORR_TRANS
#define STRARG_LEN              32                                              // Maximum length of a string argument, also length of the calibration string
#define FREQ_DRV_LVL_CHANGE     100000000UL                                     // Frequency at which main PLL chip RF drive level is changed from low to high drive level
#define DRV_LVL_LOW             12                                              // Main PLL RF drive level low value, min: 0, max: 63
#define DRV_LVL_HIGH            63                                              // Main PLL RF drive level high value, min: 0, max: 63


/*Section: Function prototypes: ************************************************/

extern void set_state_to_default(volatile Instrument_State_t *state, volatile Instrument_Error_t *error);


#ifdef	__cplusplus
}
#endif

#endif	/* APP_INSTR_STATE_H */


/*******************************************************************************
 End of File
 */