/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app_spi.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file implements the RTPOS application for sending data over the
 * SPI bus.
 * 
 * *****************************************************************************/

/*Section: Includes ************************************************************/

#include "app_spi.h"


/*Section: Function implementations ********************************************/

static void set_cs_line(SPI_Addr_t addr)
{
    switch(addr)
    {
        case SPI_ADDR_VOID:
            nCS_Ref_PLL_Set();
            nCS_PLL_Set();
            nCS_DDS_Set();
            nCS_DAC_Set();
            DDS_Update_Clear();
            break;
        case SPI_ADDR_MPLL:
            nCS_PLL_Clear();
            break;
        case SPI_ADDR_DDS:
            nCS_DDS_Clear();
            break;
        case SPI_ADDR_DDSUPDATE:
            DDS_Update_Set();
            break;
        case SPI_ADDR_REFPLL:
            nCS_Ref_PLL_Clear();
            break;
        case SPI_ADDR_DAC:
            nCS_DAC_Clear();
            break;
        default:
            break;
    }
}


/*******************************************************
 * The following functions are used for feeding data 
 * into spi_queue, which sends the data over the SPI bus,
 * and to set and clear the chip select lines.
 *******************************************************/

void spi_init(void)
{
    set_cs_line(SPI_ADDR_VOID);
}


void spi_send_reg(const uint8_t *buffer, unsigned int len, SPI_Addr_t addr)
{
    if(len == 0)
        return;
    SPI_Queue_Data_t new_message;
    new_message.addr = addr;
    for(int i = len - 1; i >= 0; i--)                                                // Transmit data MSB first (the SAM E54 is little-endian)
    {
        new_message.data = *(buffer + i);
        xQueueSendToBackFromISR(spi_queue, &new_message, NULL);
    }
}


void clr_chip_select(void)                                                      // Clear all chip select lines; must be called when no more data is to be sent to a paticular chip
{
    SPI_Queue_Data_t new_message;
    new_message.addr = SPI_ADDR_VOID;
    new_message.data = 0;
    xQueueSendToBackFromISR(spi_queue, &new_message, NULL);
}


/*******************************************************
 * The following function implements the APP_SPI_Tasks 
 * function which sends the data in spi_queue over the 
 * SPI bus.
 *******************************************************/

void APP_SPI_Tasks(void)
{
    SPI_Queue_Data_t message;
    static SPI_Addr_t old_addr = SPI_ADDR_VOID;
    
    xQueueReceive(spi_queue, &message, portMAX_DELAY);
    if(message.addr != old_addr)
    {
        set_cs_line(message.addr);
        blocking_delay(4);
    }
    if( (message.addr != SPI_ADDR_VOID) && (message.addr != SPI_ADDR_DDSUPDATE) )
        SERCOM0_SPI_Write(&message.data, 1);
    old_addr = message.addr;
}


/*******************************************************************************
 End of File
 */