/*******************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 *******************************************************************************
 * 
 * File: app_system_operation.c
 * 
 * Version: 1.0
 * Created: 16.07.2023
 * 
 * Author: Mario Hellmich <mario.hellmich@web.de>
 * Project: RF Synthesizer MK III
 * 
 * Summary: This file implements high level control routines for the synthesizer 
 * hardware.
 * 
 * *****************************************************************************/

/*Section: Includes ************************************************************/

#include "app_system_operation.h"


/*Section: Function implementation *********************************************/

/*******************************************************
 * The following functions and file scope variables 
 * handle the level setting of the generator.
 *******************************************************/

#define pow2(x) (1ULL << (x))                                                   // Powers of two, x = 0, 1, 2, ...

static const unsigned int db_step    = 3;                                       // Attenuator step in dB
const unsigned int att_nstages       = 4;                                       // Number of attenuator stages
static const unsigned int att_nsteps = pow2(att_nstages) - 1;                   // Number of attenuator steps

const float nominal_lvl_pre_att      = 0;                                       // Nominal level before attenuator in dBm
static const float intercept_lvl     = 50;                                      // Level intercept point in dBm of level detector
static const float slope_lvl         = 25;                                      // Level slope in mV/dB of level detector
static const float vref              = 2500;                                    // DAC reference voltage in mV, 2500 mV = 2.5 V
static const unsigned int dac_nbits  = 12;                                      // Number of bits of DAC


static size_t get_left_corr_idx(uint32_t freq)                                  // Get index of the nearest level calibration point to the left of freq
{
    size_t corr_idx;
    if(freq < FREQ_CORR_TRANS)
        corr_idx = (freq - FREQ_MIN) / FREQ_CORR_SPACING_LO;
    else
        corr_idx = ((freq - FREQ_CORR_TRANS) / FREQ_CORR_SPACING_HI) + ((FREQ_CORR_TRANS - FREQ_MIN) / FREQ_CORR_SPACING_LO) + 1;
    return corr_idx;
}


static uint32_t get_corr_freq(size_t idx)                                       // Get frequency corresponding to the level calibration point with index idx
{
    uint32_t freq;
    const static int nlow = ((FREQ_CORR_TRANS - FREQ_MIN) / FREQ_CORR_SPACING_LO) + 1;
    if(idx < nlow)
        freq = FREQ_MIN + (idx * FREQ_CORR_SPACING_LO);
    else
        freq = FREQ_CORR_TRANS + ((idx - nlow) * FREQ_CORR_SPACING_HI);
    return freq;
}


static float get_lvlcorr(uint32_t freq, const float *caldata)                   // Get level correction value in dB by linear interpolation between two calibration points
{
    size_t left_idx = get_left_corr_idx(freq);
    size_t right_idx;
    if(freq < FREQ_MAX)
        right_idx = left_idx + 1;
    else
        right_idx = left_idx;
    float left_val = *(caldata + left_idx);
    float right_val = *(caldata + right_idx);
    uint32_t left_freq = get_corr_freq(left_idx);
    uint32_t right_freq = get_corr_freq(right_idx);
    uint32_t delta_f;
    if(freq < FREQ_CORR_TRANS)
        delta_f = FREQ_CORR_SPACING_LO;
    else
        delta_f = FREQ_CORR_SPACING_HI;
    // Correction value, linear interpolation between left_val and right_val
    float corr;
    if(right_idx != left_idx)
        corr = (left_val * (right_freq - freq) + right_val * (freq - left_freq)) / delta_f;
    else
        corr = left_val;
    return corr;
}


float get_lvl_corr_val(uint32_t freq, unsigned int att_stage)
{
    float retval;
    switch(att_stage)
    {
        case 0:
            retval = *(caldata_ram.lvlcorr.sig_path + get_left_corr_idx(freq));
            break;
        case 1:
            retval = *(caldata_ram.lvlcorr.step_3db + get_left_corr_idx(freq));
            break;
        case 2:
            retval = *(caldata_ram.lvlcorr.step_6db + get_left_corr_idx(freq));
            break;
        case 3:
            retval = *(caldata_ram.lvlcorr.step_12db + get_left_corr_idx(freq));
            break;
        case 4:
            retval = *(caldata_ram.lvlcorr.step_24db + get_left_corr_idx(freq));
            break;
        default:
            retval = 0;
    }
    return retval;
}


void set_lvl_corr_val(uint32_t freq, unsigned int att_stage, float val)
{
    switch(att_stage)
    {
        case 0:
            caldata_ram.lvlcorr.sig_path[get_left_corr_idx(freq)] = val;
            break;
        case 1:
            caldata_ram.lvlcorr.step_3db[get_left_corr_idx(freq)] = val;
            break;
        case 2:
            caldata_ram.lvlcorr.step_6db[get_left_corr_idx(freq)] = val;
            break;
        case 3:
            caldata_ram.lvlcorr.step_12db[get_left_corr_idx(freq)] = val;
            break;
        case 4:
            caldata_ram.lvlcorr.step_24db[get_left_corr_idx(freq)] = val;
            break;
        default:
            break;
    }
}


static void lvl_write(uint64_t freq, float new_lvl)                             // Set up hardware for new level new_lvl, and apply level correcton at freq (this function does NOT set frequency)
{   
    unsigned int atten;                                                         // Number of attenuator steps to be set
    if(nominal_lvl_pre_att - new_lvl > 0)                                       // Does the attenuator need to be switched in?
    {
        atten = (nominal_lvl_pre_att - new_lvl) / db_step;
        if(atten >= att_nsteps)
            atten = att_nsteps;
    }
    else
        atten = 0;
    float lvl_corr = get_lvlcorr(freq, caldata_ram.lvlcorr.sig_path);           // Level correction value for signal path and attenuator
    if(atten & (1<<0))
        lvl_corr += get_lvlcorr(freq, caldata_ram.lvlcorr.step_3db);
    if(atten & (1<<1))
        lvl_corr += get_lvlcorr(freq, caldata_ram.lvlcorr.step_6db);
    if(atten & (1<<2))
        lvl_corr += get_lvlcorr(freq, caldata_ram.lvlcorr.step_12db);
    if(atten & (1<<3))
        lvl_corr += get_lvlcorr(freq, caldata_ram.lvlcorr.step_24db);
    float lvl_pre_atten = new_lvl + lvl_corr + (atten * db_step);               // Calculate actual level before attenuator
    uint16_t dac_val;                                                           // Level DAC value
    if(intercept_lvl - lvl_pre_atten >= 0)
        dac_val = roundf( ( (slope_lvl + caldata_ram.slopecorr) * pow2(dac_nbits) / vref ) * (intercept_lvl - lvl_pre_atten) );
    else
        dac_val = 0;
    dac_write(dac_val);                                                         // Set hardware according to the calculated values
    atten_write(atten);
}


/*******************************************************
 * The following functions and file scope variables 
 * handle the frequency setting of the synthesizer.
 *******************************************************/

static const uint32_t f_vco_min         = 3600000000;                           // Minimum VCO frequency
static const unsigned int pre           = 2;                                    // Prescaler division between VCO and N divider, set to 2 for performance reasons
static const unsigned int n_div_min     = 9;                                    // Minimum PLL integer divider, set to 9 as per datasheet for integer-n operation
static const unsigned int pll_ref_mult  = 1;                                    // Reference path doubler of the LMX2582; set to 2 when enabled
static const uint32_t hfilt_range_end[] = {36000000, 65000000, 117000000, 211000000, 380000000, 685000000, 1230000000, 2215000000, 4200000000};


static unsigned int mpll_chdivreg(unsigned int div)
{
    unsigned int retval;
    switch(div)
    {
        case 1:
            retval = 0;
            break;
        case 2:
            retval = 1;
            break;
        case 4:
            retval = 2;
            break;
        case 6:
            retval = 4;
            break;
        case 8:
            retval = 8;
            break;
        default:
            retval = 0;
            break;
    }
    return retval;
}


static void get_dds_synth_params(long double dds_fout, uint32_t *dds_tunewd)
{
    long double f_dds_sysclk_corr;                                              // DDS sysclk corrected by frequency calibration value
    if(state.extref == true)
        f_dds_sysclk_corr = f_dds_sysclk;
    else
        f_dds_sysclk_corr = ( (long double)f_dds_sysclk ) * (1 + (long double)caldata_ram.frqcorr/f_int_ref);
    *dds_tunewd = lroundl((dds_fout * pow2(32)) / f_dds_sysclk_corr);           // Calculate and store DDS tuning word 
}


static void get_pll_synth_params(uint32_t new_freq, bool rf_on, struct mpll_s *params, uint32_t *dds_tunewd, unsigned int *hfiltnumber)
{
    params->outa_mux = false;                                                   // Initialize struct member to avoid compiler warning
    if(new_freq >= FREQ_BAND_CHANGE && rf_on == true)                           // Is state.rfon set and is main PLL used for output signal generation?
        params->rfon = true;                                                    // Then enable main PLL output 
    else
    {
        params->rfon = false;                                                   // Else disable main PLL output
        new_freq = FREQ_BAND_CHANGE;                                            // and set main PLL synthesizer to minimum frequency
    }
    
    if(new_freq <= FREQ_DRV_LVL_CHANGE)                                         // Is new_freq lower or equal than frequency for RF drive level change of main PLL chip?
        params->drive_lvl = DRV_LVL_LOW;                                        // Then set low drive level
    else
        params->drive_lvl = DRV_LVL_HIGH;                                       // Else set high drive level
    
    /* Calculate divider settings for nominal PFD frequency */
    unsigned int seg1, seg2, seg3;                                              // Channel divider division settings
    unsigned int chdiv_tot;                                                     // Channel divider total division
    uint64_t f_vco_tgt;                                                         // VCO target frequency that would be needed when PLL is clocked exactly with nominal PFD frequency f_pfd
    unsigned int n_div;
    if( new_freq < f_vco_min )                                                  // Is channel division required?
    {        
        for( seg3 = 1; seg3 <= 8; seg3 = ( (seg3==1) ? 2 : seg3+2 ) )           // Determine channel divider settings
        {
            for( seg2 = seg3; seg2 <= 8; seg2 = ( (seg2==1) ? 2 : seg2+2 ) )
            {
                for( seg1 = 2; seg1 <= 3; seg1++ )
                {
                    chdiv_tot = seg1 * seg2 * seg3;                             // Calculate the resulting total channel division
                    f_vco_tgt = (uint64_t)new_freq * chdiv_tot;                 // and the resulting VCO frequency
                    if(f_vco_tgt >= f_vco_min)                                  // Is the resulting VCO frequency greater than the minimum VCO frequency?
                    {
                        n_div = ceill( (long double)f_vco_tgt / (f_pfd * pre) );// Then calculate corresponding PLL integer divider
                        if(n_div >= n_div_min)                                  // Is current PLL integer division greater than allowable minimum?
                        {
                            params->outa_mux = false;                           // Then route signal from channel divider output to RF output
                            goto final;                                         // and jump out of loop
                        }
                    }
                }
            }
        }
    }
    else                                                                        // Channel division not required
    {
        seg1 = seg2 = seg3 = 1;
        f_vco_tgt = new_freq;
        params->outa_mux = true;                                                // Route VCO signal directly to RF output
        goto final;
    }
    
    /* Continue here when channel divider setting calculation is finished */
    final:
    if(params->outa_mux == false)                                               // Is channel division required?
    {
        params->chdiv_seg1 = ( seg1 == 3 ? 1 : 0 );                             // Set channel divider segment 1 division
        params->chdiv_seg2 = mpll_chdivreg(seg2);                               // Set channel divider segment 2 division
        params->chdiv_seg3 = mpll_chdivreg(seg3);                               // Set channel divider segment 3 division
        if(seg3 > 1)                                                            // Is division by channel divider segment 3 required (then also segments 1 and 2 will be required)?
            params->chdiv_seg_sel = (1<<2);                                     // Include channel divider segments 1, 2 and 3
        else if(seg2 > 1)                                                       // Is division from channel divider segment 2 required (then also segment 1 will be required)?
            params->chdiv_seg_sel = (1<<1);                                     // Include channel divider segments 1 and 2
        else                                                                    // Channel divider segments 2 and 3 are not required, hence only segment 1 is required
            params->chdiv_seg_sel = (1<<0);                                     // Include only channel divider segment 1
    }                
    
    /* Calculate PLL divider and DDS tuning word */
    n_div = ceill( (long double)f_vco_tgt / (f_pfd * pre) );                    // PLL integer divider
    long double f_dds = (long double)f_vco_tgt / (pll_ref_mult * n_div * pre);
    get_dds_synth_params(f_dds, dds_tunewd);                                    // Calculate DDS tuning word and store it in dds_tunewd
    params->pll_int = n_div;
    
    /* Determine harmonic filter setting */
    unsigned int num = 0;                                                       // Harmonic filter to be switched into the signal path for the set frequency
    while(new_freq > hfilt_range_end[num])
        num++;
    *hfiltnumber = num;
}


/*******************************************************
 * The following function is to be called by the RTOS 
 * task xAPP_Syscontrol_Tasks, which normally blocks 
 * unless it is woken by set_output(). Wake-up can be 
 * either from another task or from an ISR in sweep 
 * mode.
 *******************************************************/

static volatile SYS_Output_Setting_t sys_output_setting = {.freq = 0, .lvl = 0, .rf_on = false, .modified = false};

void SYS_Operation_Tasks(void)
{
    /* Block task xAPP_Syscontrol_Tasks here for 50 ms and wait for a notification */
    ulTaskNotifyTake(pdTRUE, 50 / portTICK_PERIOD_MS);
    
    uint32_t new_freq;
    float new_lvl;
    bool rf_on;
    bool update_hardware;
    
    taskENTER_CRITICAL();                                                       // Disable interrupts and context switches
        if(sys_output_setting.modified == true)
        {
            new_freq = sys_output_setting.freq;
            new_lvl  = sys_output_setting.lvl;
            rf_on    = sys_output_setting.rf_on;
            sys_output_setting.modified = false;
            update_hardware = true;
        }
        else
            update_hardware = false;
    taskEXIT_CRITICAL();                                                        // Enable interrupts and context switches
    
    if(update_hardware == true)
    {
        struct mpll_s params;                                                   // Container for main PLL chip setup data
        uint32_t dds_tunewd;                                                    // DDS tuning word 
        unsigned int hfiltn;                                                    // Harmonic filter number

        if(rf_on == true)
        {
            /* RF output is selected on */
            if(new_freq >= FREQ_BAND_CHANGE)
            {
                /* Output signal will be generated by PLL synthesizer */
                get_pll_synth_params(new_freq, true, &params, &dds_tunewd, &hfiltn); // Calculate params for main PLL even when the output will be generated by DDS in order to set params.rfon appropriately
                dds_write_tunewd(dds_tunewd);
                bandselect_pllout(true);                                        // Route main PLL synthesizer to output
                hfilter_write(hfiltn);
                blocking_delay(50);                                             // Wait a bit to account for the DDS frequency change propagation delay
                mpll_write(&params);
            }
            else
            {
                /* Output signal will be generated by DDS synthesizer */
                get_pll_synth_params(FREQ_BAND_CHANGE, false, &params, &dds_tunewd, &hfiltn);
                hfilter_write(0);
                mpll_write(&params);                                            // Write main PLL parameters in order to select output off
                bandselect_pllout(false);                                       // Route DDS synthesizer to output
                get_dds_synth_params(new_freq, &dds_tunewd);                    // Update dds_tunewd for required DDS output frequency
                dds_write_tunewd(dds_tunewd);
            }
        }
        else
        {
            /* RF output is selected off */
            hfilter_write(0);
            bandselect_pllout(true);                                            // If the RF output is selected off, route PLL chip to output and disable PLL output in order to power down the LF-Out signal
            get_pll_synth_params(FREQ_BAND_CHANGE, false, &params, &dds_tunewd, &hfiltn);
            get_dds_synth_params(f_dds_ref, &dds_tunewd);
            mpll_write(&params);                                                // Write main PLL parameters in order to select output off
            dds_write_tunewd(dds_tunewd);                                       // Set DDS to f_pdf (100 MHz) to achieve additional attenuation by harmonic filter
            new_lvl = LVL_MIN;                                                  // Set level to minimum in order to set attenuator to maximum attenuation; this helps to reduce leakage to the output
        }
        /* Set level to correct for new_freq, or set minimum level when RF output is selected off */
        lvl_write(new_freq, new_lvl);
    }
}


/*******************************************************
 * The following function schedules a setting of both 
 * the output frequency and the corrected level of the 
 * synthesizer, and unblocks xAPP_Syscontrol_Tasks,
 * which then performs the actual setting of the 
 * synthesizer hardware.
 *******************************************************/

static void set_output(uint32_t new_freq, float new_lvl, bool new_rf_state)     // Schedule a setting of the generator to frequency new_freq, level new_lvl, and new_rf_state
{
    UBaseType_t uxSavedInterruptStatus = taskENTER_CRITICAL_FROM_ISR();
        sys_output_setting.freq = new_freq;
        sys_output_setting.lvl = new_lvl;
        sys_output_setting.rf_on = new_rf_state;
        sys_output_setting.modified = true;
    taskEXIT_CRITICAL_FROM_ISR(uxSavedInterruptStatus);
    /* Notify and unblock xAPP_Syscontrol_Tasks to handle synthesizer output signal setting */
    BaseType_t xHigherPriorityTaskWoken = pdFALSE;
    vTaskNotifyGiveFromISR(xAPP_Syscontrol_Tasks, pdFALSE);
    portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}


/*******************************************************
 * The following functions and file scope variables 
 * implement the frequency sweep functionality.
 *******************************************************/

static struct {
    uint64_t nextfreq;
    unsigned int nextn;
} sweep;

static void sweep_timer_callback(TC_COMPARE_STATUS status, unsigned int context)
{
    if(++sweep.nextn < state.npoints)
    {
        long double delta_f = state.stop_freq - state.start_freq;
        sweep.nextfreq = (uint64_t)(
                                        0.5 + (long double)state.start_freq + ( ( (long double)sweep.nextn / (state.npoints - 1) ) * delta_f )
                                   );
    }
    else if(state.single_sweep == false)
    {
        /* Continuous sweep mode; sweep has reached maximum frequency */
        sweep.nextfreq = state.start_freq;
        sweep.nextn = 0;        
    }
    else
    {
        /* Single sweep mode; sweep has reached maximum frequency */
        TC0_CompareCommandSet(TC_COMMAND_STOP);
        return;
    }
    /* Schedule frequency and level setting */
    set_output(sweep.nextfreq, state.level, state.rfon);
}


void Sweeptimer_Init(void)
{
    /* Register sweep timer compare match callback function with library; this 
     * function will be called by the compare match interrupt handler */
    TC0_CompareCallbackRegister(sweep_timer_callback, 0);
    /* Start peripheral, software stop and reset the timer counter */
    TC0_CompareStart();
    TC0_CompareCommandSet(TC_COMMAND_STOP);
    TC0_Compare32bitCounterSet(0);
}


/*******************************************************
 * Externally accessible functions for high level 
 * operation of the generator. To be called mainly by 
 * the command parser.
 *******************************************************/

#define MSEC_TO_COMPARE0_VAL(x)     (6000U * (x))

void synthesizer_update(void)                                                   // Should be called when state has been changed to set the synthesizer hardware accordingly
{
    /* Stop sweep timer and reset TC0 32-bit counter */
    TC0_CompareCommandSet(TC_COMMAND_STOP);
    if(state.cal == false)
    {
        if(state.sweep == false)
        {
            /* Instrument is in CW mode */
            set_output(state.cw_freq, state.level, state.rfon);
        }
        else
        {
            /* Instrument is in sweep mode */
            unsigned int dwell_msec = lroundf(state.dwell);
            TC0_Compare32bitPeriodSet(MSEC_TO_COMPARE0_VAL(dwell_msec));
            TC0_Compare32bitCounterSet(0);
            TC0_CompareCommandSet(TC_COMMAND_FORCE_UPDATE);
            sweep.nextfreq = state.start_freq;
            sweep.nextn = 0;
            set_output(sweep.nextfreq, state.level, state.rfon);
            /* Start sweep timer */
            TC0_CompareCommandSet(TC_COMMAND_START_RETRIGGER);
        }
    }
    else
    {
        /* Instrument is in calibration mode */
        switch(state.cal_mode)
        {
            case CALMODE_NORMAL: ;
                float level;
                if(state.cal_att_stage == 0)
                    level = nominal_lvl_pre_att;
                else
                    level = nominal_lvl_pre_att - (db_step * pow2(state.cal_att_stage - 1));
                set_output(state.cw_freq, level, true);
                break;
            
            case CALMODE_SLOPE_HIGH:
                set_output(state.cw_freq, nominal_lvl_pre_att + SLOPE_CORR_LVL_DIFF, true);
                break;
                
            case CALMODE_SLOPE_LOW:
                set_output(state.cw_freq, nominal_lvl_pre_att, true);
                break;
                
            default:
                break;
        }
    }
}


void ref_update(void)                                                           // Should be called when the settings pertaining to the reference frequency input have been modified to set the synthesizer hardware accordingly
{
    if(state.extref == false)
    {
        /* Use internal reference and matching PLL multiplier */
        refpll_set_ext_ref(false);
        refpll_write_mult( (uint32_t)(f_dds_ref / f_int_ref) );
    }
    else
    {
        /* Use external reference */
        refpll_set_ext_ref(true);
        refpll_write_mult(state.refmult);
    }
    synthesizer_update();
}


void set_sweep_mode(bool swp_state)                                             // Set or disable sweep mode
{
    if(swp_state == true)
    {
        state.sweep = true;
        synthesizer_update();
    }
    else
    {
        state.sweep = false;
        synthesizer_update();
    }
}


bool set_cw_freq_hz(uint64_t freq)                                              // Set CW frequency to freq (in Hz))
{
    if(freq <= FREQ_MAX && freq >= FREQ_MIN)
    {
        state.cw_freq = freq;
        if(state.sweep == false)
            synthesizer_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


bool set_lvl_db(float lvl)                                                      // Set level to lvl (in dBm)
{
    if(lvl <= LVL_MAX && lvl >= LVL_MIN)
    {
        state.level = lvl;
        if(state.sweep == false)
            synthesizer_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


bool set_start_frq_hz(uint64_t freq)                                            // Set sweep start frequency to freq (in Hz))
{
    if(freq <= FREQ_MAX && freq >= FREQ_MIN)
    {
        state.start_freq = freq;
        if(freq > state.stop_freq)
            state.stop_freq = freq;
        if(state.sweep == true)
            synthesizer_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


bool set_stop_frq_hz(uint64_t freq)                                             // Set sweep stop frequency to freq (in Hz))
{
    if(freq <= FREQ_MAX && freq >= FREQ_MIN)
    {
        state.stop_freq = freq;
        if(freq < state.start_freq)
            state.start_freq = freq;
        if(state.sweep == true)
            synthesizer_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


bool set_dwell_time_ms(float time)                                              // Set sweep dwell time to time (in ms))
{
    if(time <= DWELL_MAX && time >= DWELL_MIN)
    {
        state.dwell = time;
        if(state.sweep == true)
        {
            unsigned int dwell_msec = lroundf(state.dwell);
            TC0_Compare32bitPeriodSet(MSEC_TO_COMPARE0_VAL(dwell_msec));
            TC0_Compare32bitCounterSet(0);
            TC0_CompareCommandSet(TC_COMMAND_FORCE_UPDATE);
        }
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


bool set_npoints(unsigned int npoints)                                          // Set number of sweep points to npoints
{
    if(npoints <= NPOINTS_MAX && npoints >= NPOINTS_MIN)
    {
        state.npoints = npoints;
        if(state.sweep == true)
            synthesizer_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


void set_sweep_single(bool swp_single)                                          // Set or disable single sweep mode
{
    if(swp_single == true)
        state.single_sweep = true;
    else
        state.single_sweep = false;
}


void set_extref(bool ref_state)                                                 // Set or disable external reference frequency input
{
    if(ref_state == true)
        state.extref = true;
    else
        state.extref = false;
    ref_update();
}


bool set_ref_frq_hz(uint64_t reffreq)                                           // Set external reference input frequency to freq (in Hz))
{
    
    if(reffreq <= REFMULT_MAX && reffreq >= REFMULT_MIN)
    {
        state.refmult = (uint32_t)(f_pfd / reffreq);
        ref_update();
        return EXIT_SUCCESS;
    }
    else
        return EXIT_FAILURE;
}


void system_reset(void)                                                         // Reload calibration data to SRAM, load state from memory 0 or set to default, reset and reconfigure all hardware
{
    APP_Initialize();
}


/*******************************************************
 * A function to dump all calibration data over the
 * USB CDC data link. To be called by the command
 * parser.
 *******************************************************/

#define BUFLEN          128

static void cdc_send_str(const char *str)                                       // Send a string str over the USB CDC data link
{
    for(unsigned int i = 0; str[i] != '\0'; i++)
        xQueueSendToBack(cdc_tx_queue, str + i, portMAX_DELAY);
}


void dump_cal_data(void)                                                        // Dump all calibration data from RAM over the USB CDC data link
{
    char buf[BUFLEN];
    strcpy(buf, "******************************* CALIBRATION DATA ******************************\r\n");
    cdc_send_str(buf);
    sprintf(buf, "Calibration String: %s\r\n", caldata_ram.calstring);
    cdc_send_str(buf);
    strcpy(buf, "*******************************************************************************\r\n");
    cdc_send_str(buf);
    for(unsigned int att_stage = 0; att_stage <= att_nstages; att_stage++)
    {
        sprintf(buf, "****** Attenuator Stage %u *****************************************************\r\n", att_stage);
        cdc_send_str(buf);
        for(unsigned int i = 0; i < N_LVL_CORR_POINTS; i++)
        {
            uint32_t corr_freq = get_corr_freq(i);
            sprintf(buf, "Freq %lu Hz Value: %.6f\r\n", corr_freq, get_lvl_corr_val(corr_freq, att_stage));
            cdc_send_str(buf);
        }
    }
    sprintf(buf, "****** Detector Slope *********************************************************\r\nValue: %.6f\r\n", caldata_ram.slopecorr);
    cdc_send_str(buf);
    sprintf(buf, "****** Timebase ***************************************************************\r\nValue: %.8f\r\n", caldata_ram.frqcorr);
    cdc_send_str(buf);
}


/*******************************************************
 * Continuously running built-in self test and error
 * messages.
 *******************************************************/

void selftest(void)
{
    if((state.rfon == false) || (state.sweep == true) )
        error.ALC_UNLEVELED = false;
    else
        error.ALC_UNLEVELED = Unlvld_Get();
   
    if( (state.sweep == true) || (state.cw_freq < FREQ_BAND_CHANGE) )
        error.MAINPLL_UNLK = false;
    else
        error.MAINPLL_UNLK  = !PLL_Lock_Get();
     
    error.REFPLL_UNLK   = !Ref_Lock_Get();
    
    if(error.errstatus == 0)
    {
        FP_LED_red_Clear();
        FP_LED_grn_Set();
        LED2_Clear();        
    }
    else
    {
        FP_LED_red_Set();
        FP_LED_grn_Clear();
        LED2_Set();
    }
    static int count = 0;
    if(count++ > (1<<8))
    {
        LED1_Toggle();
        count = 0;
    }
}


void print_errmsg(void)
{
    char buf[BUFLEN];
    strcpy(buf, "******************************* Current Errors ********************************\r\n");
    cdc_send_str(buf);
    if(error.errstatus == 0)
    {
        strcpy(buf, "No pending errors");
        cdc_send_str(buf);
    }
    else
    {
        if(error.ARGUMENT_OUT_OF_RANGE == true)
        {
            strcpy(buf, "Argument out of range\r\n");
            cdc_send_str(buf);
        }
        if(error.MAINPLL_UNLK == true)
        {
            strcpy(buf, "Main PLL unlock\r\n");
            cdc_send_str(buf);
        }
        if(error.REFPLL_UNLK == true)
        {
            strcpy(buf, "Reference PLL unlock\r\n");
            cdc_send_str(buf);
        }
        if(error.ALC_UNLEVELED == true)
        {
            strcpy(buf, "ALC error, output unleveled\r\n");
            cdc_send_str(buf);
        }
        if(error.CAL_PROTECTED == true)
        {
            strcpy(buf, "Calibration data protected\r\n");
            cdc_send_str(buf);
        }
        if(error.EEPROM_CALDATA_CHKSUM == true)
        {
            strcpy(buf, "No or invalid calibration data, default loaded\r\n");
            cdc_send_str(buf);
        }
        if(error.EEPROM_ERROR == true)
        {
            strcpy(buf, "EEPROM error\r\n");
            cdc_send_str(buf);
        }
    }
}


/*******************************************************
 * Initialize hardware and set synthesizer to startup
 * state.
 *******************************************************/

void HW_Initialize(void)
{
    FP_LED_red_Clear();
    FP_LED_grn_Clear();
    synthesizer_power_down(false);
    
    /* To be called at system startup */
    vTaskDelay(250 / portTICK_PERIOD_MS);
    spi_init();
    
    refpll_init();
    vTaskDelay(100 / portTICK_PERIOD_MS);
    
    dds_init();
    vTaskDelay(100 / portTICK_PERIOD_MS);
    
    mpll_init();
    vTaskDelay(100 / portTICK_PERIOD_MS);
        
    FP_LED_grn_Set();                                                           // Turn on green front panel LED
}


/*******************************************************************************
 End of File
 */