% File 'freq_flatness.m'
% 
% For GNU Octave
% --------------------------------------------------------------------------------------------------
% --------------------------------------------------------------------------------------------------
% 
% Octave script to perform a level flatness measurement of the Mk III Signal Generator.
%
% This script is written to work with a Rohde & Schwarz NRVD power meter
% with power sensor connected to channel A, and controlled by a Prologix 
% GPIB/USB controller.
% 
% Author: Dr. Mario Hellmich <mario.hellmich@web.de>
% Date:   12.07.2023
% 
% Uses the Octave Forge 'instrument-control' package 
%
% --------------------------------------------------------------------------------------------------
% --------------------------------------------------------------------------------------------------



1;                                                                           				   		% Prevent Octave from thinking this is a function file
clear;                                                                          					% Delete all
clc;																								% Clear screen
pkg load "instrument-control"																		% Load instrument control package
pause(0.2);


% --------------------------------------------------------------------------------------------------
% ----- Com Setup ----------------------------------------------------------------------------------

global instr = serialport("COM9", 9600);															% Open virtual COM port for the signal generator (must be the port of the signal generator)
global meter = serialport("COM6", 9600);															% Open virtual COM port for the power meter (must be the port of the GPIB controller connected to the R&S NRVD)

set(instr, 'DataBits', 8);
set(instr, 'parity', 'n');
set(instr, 'stopbits', 1);
set(instr, 'timeout', 10);
configureTerminator(instr, 'lf', 'lf');

set(meter, 'DataBits', 8);
set(meter, 'parity', 'n');
set(meter, 'stopbits', 1);
set(meter, 'timeout', 10);
configureTerminator(meter, 'lf', 'lf');

flush(instr);																						% Flush port
flush(meter);																						% Flush port																						% Flush port


% --------------------------------------------------------------------------------------------------
% ----- Constants ----------------------------------------------------------------------------------

address = 13;																						% GPIB address of my R&S NRVD

global fmin   = 1e6;																				% Minimum frequency (Hz)
global fmax   = 4.2e9																				% Maximum frequency (Hz)
global npt    = 201;																				% Number of frequency test points

global lvlmin = -45;																				% Minimum level (dBm)
global lvlmax = 10; 																				% Maximum level (dBm)
global lvlstp = 5;																					% Level step (dB)


% --------------------------------------------------------------------------------------------------
% ----- Utility functions --------------------------------------------------------------------------

function int_array = readtolf(ser_handle)															% Read characters from a serial port with handle ser_handle until a line-feed character occurs
	int_array = [];
	do
		nextchar = fread(ser_handle, 1);
		int_array(end + 1) = nextchar;
	until(char(nextchar) == "\n")
endfunction


function retval = readnumtolf(ser_handle)
	retval = str2num(char(readtolf(ser_handle)));
endfunction


% --------------------------------------------------------------------------------------------------
% ----- Setup --------------------------------------------------------------------------------------

fwrite(meter, "++mode 1\n");																		% Set the GPIB controller to controller mode
%fwrite(meter, ["++addr ", num2str(address), " \r\n"]);												% Set NRVD GPIB address
%fwrite(meter, "++auto 1\n");																		% Set GPIB controller to auto listen mode

fwrite(meter, "*RST;*CLS;\n");																		% NRVD reset and clear status
fwrite(meter, "INP:SEL \"B\"\n");																	% Select channel B
fwrite(meter, "SENS:POW:UNIT DBM\n");																% Select power measurement and unit dBm
fwrite(meter, "DISP:ANN:POW:NRES 5\n");																% Select 5 digits resolution
fwrite(meter, "DISP:ANN:BARG ON\n");																% Switch on bargraph on NRVD
fwrite(meter, "DISP:ANN:BARG:AUTO\n");																% Select bargraph autoranging
fwrite(meter, "SENS:CORR:FREF:STAT ON\n");															% Select frequency response correction on
fwrite(meter, "DIAG:TEST2 OFF\n");																	% No default string when output buffer is empty

fwrite(instr, "rst\r\n");																			% Reset signal generator
pause(4);
fwrite(instr, "ref:ext 1\r\n");																		% Use an external frequency reference


% --------------------------------------------------------------------------------------------------
% ----- Measurement --------------------------------------------------------------------------------

freq = linspace(fmin, fmax, npt);																	% Frequency points
lvl  = [lvlmin : lvlstp : lvlmax];																	% Levels to measure

disp("Level flatness measurement:\n");

for l = 1:numel(lvl)
	fwrite(instr, ["cw:lvl ", num2str(lvl(l)), " dBm\r\n"]);										% Set signal generator to new level
	for f = 1:numel(freq)
		curr_freq = round(freq(f));
		fwrite(instr, ["cw:frq ", num2str(curr_freq), " Hz\r\n"]);									% Set signal generator to new frequency
		fwrite(meter, ["SENS:CORR:FREF ", num2str(curr_freq), " Hz\n"]);							% Set NRVD power sensor frequency correction
		fwrite(meter, "*TRG;*WAI\n");																% Trigger NRVD
		result(l, f) = readnumtolf(meter);															% Read and store result
		printf("Measuring at %e Hz with level set to %d dBm, result %e dBm...\n", freq(f), lvl(l), result(l, f));
	endfor
endfor


% --------------------------------------------------------------------------------------------------
% ----- Plotting -----------------------------------------------------------------------------------

hold on;
plot(freq, result, "linewidth", 2);
set(gca, "linewidth", 2, "fontsize", 12);
title("Level Flatness", "fontsize", 16);
xlabel("Frequency/Hz");
ylabel("Level/dBm");
xlim([fmin fmax]);
ylim([(lvlmin - 5) (lvlmax + 5)])
for l = lvl
	text(2e9, l, [num2str(l), " dBm"]);
endfor
pbaspect([4, 3, 3]);
box on;
grid on;

print -dpng "-S1800,1350" "level_flatness.png"

% --------------------------------------------------------------------------------------------------
% End of File
% 